#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

from pyomo.environ import (
    ConcreteModel,
    Var,
    value,
    Constraint,
    Objective,
    TransformationFactory,
    units as pyunits,
)
import idaes.logger as idaeslog
from pyomo.network import Arc

from idaes.core import FlowsheetBlock, UnitModelCostingBlock
from watertap.core.solvers import get_solver
from idaes.core.util.initialization import propagate_state
from idaes.core.util.model_statistics import report_statistics
from idaes.models.unit_models import Feed, Product, Separator
from pandas import DataFrame
import idaes.core.util.scaling as iscale
from watertap.core.util.initialization import check_dof, check_solve
from watertap.unit_models.electrodialysis_1D import Electrodialysis1D

from watertap.costing import WaterTAPCosting
from watertap.property_models.multicomp_aq_sol_prop_pack import MCASParameterBlock


# Set up logger
_log = idaeslog.getLogger(__name__)

__author__ = "Xiangyu Bi"


def main():
    # set up solver
    solver = get_solver()
    # Simulate a fully defined operation
    m = build()
    set_operating_conditions(m)
    initialize_system(m, solver=solver)
    solve(m, solver=solver, checkpoint="solve flowsheet after initializing system")

    print("\n***---Simulation results---***")
    m.fs.EDstack.report()
    display_model_metrics(m)

    # Perform an optimization over selected variables
    initialize_system(m, solver=solver)
    optimize_system(
        m, solver=solver, checkpoint="solve flowsheet after optimizing system"
    )
    print("\n***---Optimization results---***")
    m.fs.EDstack.report()
    display_model_metrics(m)


def build():
    # ---building model---
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    ion_dict = {
        "solute_list": ["Na_+", "Cl_-"],
        "mw_data": {"H2O": 18e-3, "Na_+": 23e-3, "Cl_-": 35.5e-3},
        "elec_mobility_data": {("Liq", "Na_+"): 5.19e-8, ("Liq", "Cl_-"): 7.92e-8},
        "charge": {"Na_+": 1, "Cl_-": -1},
    }
    m.fs.properties = MCASParameterBlock(**ion_dict)
    m.fs.costing = WaterTAPCosting()
    m.fs.feed = Feed(property_package=m.fs.properties)
    m.fs.separator = Separator(
        property_package=m.fs.properties,
        outlet_list=["inlet_diluate", "inlet_concentrate"],
    )  # "inlet_diluate" and "inlet_concentrate" are two separator's outlet ports that are connected to the two inlets of the ED stack.

    # Add electrodialysis (ED) stacks
    m.fs.EDstack = Electrodialysis1D(
        property_package=m.fs.properties,
        operation_mode="Constant_Voltage",
        finite_elements=20,
    )
    m.fs.product = Product(property_package=m.fs.properties)
    m.fs.disposal = Product(property_package=m.fs.properties)

    # Touching needed variables for initialization and displaying results
    m.fs.feed.properties[0].conc_mass_phase_comp[...]
    m.fs.separator.inlet_diluate_state[0].conc_mass_phase_comp[...]
    m.fs.separator.inlet_concentrate_state[0].conc_mass_phase_comp[...]
    m.fs.product.properties[0].conc_mass_phase_comp[...]
    m.fs.disposal.properties[0].conc_mass_phase_comp[...]

    m.fs.feed.properties[0].flow_vol_phase[...]
    m.fs.separator.inlet_diluate_state[0].flow_vol_phase[...]
    m.fs.separator.inlet_concentrate_state[0].flow_vol_phase[...]
    m.fs.product.properties[0].flow_vol_phase[...]
    m.fs.disposal.properties[0].flow_vol_phase[...]

    # costing
    m.fs.EDstack.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
    m.fs.costing.cost_process()
    m.fs.costing.add_annual_water_production(
        m.fs.product.properties[0].flow_vol_phase["Liq"]
    )
    m.fs.costing.add_LCOW(m.fs.product.properties[0].flow_vol)
    m.fs.costing.add_specific_energy_consumption(
        m.fs.product.properties[0].flow_vol_phase["Liq"]
    )

    # Add two variable for reporting
    m.fs.mem_area = Var(
        initialize=1,
        bounds=(0, 1e3),
        units=pyunits.meter**2,
        doc="Total membrane area for cem (or aem) in one stack",
    )
    m.fs.product_salinity = Var(
        initialize=1, bounds=(0, 1000), units=pyunits.kg * pyunits.meter**-3
    )
    m.fs.disposal_salinity = Var(
        initialize=1, bounds=(0, 1e6), units=pyunits.kg * pyunits.meter**-3
    )
    m.fs.eq_product_salinity = Constraint(
        expr=m.fs.product_salinity
        == sum(
            m.fs.product.properties[0].conc_mass_phase_comp["Liq", j]
            for j in m.fs.properties.ion_set | m.fs.properties.solute_set
        )
    )
    m.fs.eq_disposal_salinity = Constraint(
        expr=m.fs.disposal_salinity
        == sum(
            m.fs.disposal.properties[0].conc_mass_phase_comp["Liq", j]
            for j in m.fs.properties.ion_set | m.fs.properties.solute_set
        )
    )

    m.fs.eq_mem_area = Constraint(
        expr=m.fs.mem_area
        == m.fs.EDstack.cell_width
        * m.fs.EDstack.cell_length
        * m.fs.EDstack.cell_pair_num
    )

    # Add Arcs
    m.fs.s01 = Arc(source=m.fs.feed.outlet, destination=m.fs.separator.inlet)
    m.fs.s02 = Arc(
        source=m.fs.separator.inlet_diluate, destination=m.fs.EDstack.inlet_diluate
    )
    m.fs.s03 = Arc(
        source=m.fs.separator.inlet_concentrate,
        destination=m.fs.EDstack.inlet_concentrate,
    )
    m.fs.s04 = Arc(source=m.fs.EDstack.outlet_diluate, destination=m.fs.product.inlet)
    m.fs.s05 = Arc(
        source=m.fs.EDstack.outlet_concentrate, destination=m.fs.disposal.inlet
    )
    TransformationFactory("network.expand_arcs").apply_to(m)

    # Scaling
    m.fs.properties.set_default_scaling("flow_mol_phase_comp", 1, index=("Liq", "H2O"))
    m.fs.properties.set_default_scaling(
        "flow_mol_phase_comp", 5e1, index=("Liq", "Na_+")
    )
    m.fs.properties.set_default_scaling(
        "flow_mol_phase_comp", 5e1, index=("Liq", "Cl_-")
    )
    iscale.set_scaling_factor(m.fs.EDstack.cell_width, 10)
    iscale.set_scaling_factor(m.fs.EDstack.cell_length, 10)
    iscale.calculate_scaling_factors(m)
    return m


def set_operating_conditions(m):

    solver = get_solver()

    # ---specifications---
    # Fix state variables at the origin
    m.fs.feed.properties[0].pressure.fix(101325)  # feed pressure [Pa]
    m.fs.feed.properties[0].temperature.fix(273.15 + 25)  # feed temperature [K]
    m.fs.feed.properties[0].flow_mol_phase_comp["Liq", "H2O"].fix(4.8)
    m.fs.feed.properties[0].flow_mol_phase_comp["Liq", "Na_+"].fix(1.476e-2)
    m.fs.feed.properties[0].flow_mol_phase_comp["Liq", "Cl_-"].fix(1.476e-2)

    # Fix separator's split_fraction to 0.5, i.e., equal flows into the diluate and concentrate channels
    m.fs.separator.split_fraction[0, "inlet_diluate"].fix(0.5)

    # Fix ED unit vars
    m.fs.EDstack.water_trans_number_membrane["cem"].fix(5.8)
    m.fs.EDstack.water_trans_number_membrane["aem"].fix(4.3)
    m.fs.EDstack.water_permeability_membrane["cem"].fix(2.16e-14)
    m.fs.EDstack.water_permeability_membrane["aem"].fix(1.75e-14)
    m.fs.EDstack.voltage_applied.fix(5)
    m.fs.EDstack.electrodes_resistance.fix(0)
    m.fs.EDstack.cell_pair_num.fix(100)
    m.fs.EDstack.current_utilization.fix(1)
    m.fs.EDstack.channel_height.fix(2.7e-4)
    m.fs.EDstack.membrane_areal_resistance["cem"].fix(1.89e-4)
    m.fs.EDstack.membrane_areal_resistance["aem"].fix(1.77e-4)
    m.fs.EDstack.membrane_areal_resistance_coef["cem"].fix(0)
    m.fs.EDstack.membrane_areal_resistance_coef["aem"].fix(0)
    m.fs.EDstack.cell_width.fix(0.1)
    m.fs.EDstack.cell_length.fix(0.79)
    m.fs.EDstack.membrane_thickness["aem"].fix(1.3e-4)
    m.fs.EDstack.membrane_thickness["cem"].fix(1.3e-4)
    m.fs.EDstack.solute_diffusivity_membrane["cem", "Na_+"].fix(1.8e-10)
    m.fs.EDstack.solute_diffusivity_membrane["aem", "Na_+"].fix(1.25e-10)
    m.fs.EDstack.solute_diffusivity_membrane["cem", "Cl_-"].fix(1.8e-10)
    m.fs.EDstack.solute_diffusivity_membrane["aem", "Cl_-"].fix(1.25e-10)
    m.fs.EDstack.ion_trans_number_membrane["cem", "Na_+"].fix(1)
    m.fs.EDstack.ion_trans_number_membrane["aem", "Na_+"].fix(0)
    m.fs.EDstack.ion_trans_number_membrane["cem", "Cl_-"].fix(0)
    m.fs.EDstack.ion_trans_number_membrane["aem", "Cl_-"].fix(1)
    m.fs.EDstack.spacer_porosity.fix(1)
    m.fs.EDstack.spacer_conductivity_coefficient.fix(1)

    # check zero degrees of freedom
    check_dof(m)


def solve(blk, solver=None, checkpoint=None, tee=False, fail_flag=True):
    if solver is None:
        solver = get_solver()
    results = solver.solve(blk, tee=tee)
    check_solve(results, checkpoint=checkpoint, logger=_log, fail_flag=fail_flag)
    return results


def initialize_system(m, solver=None):

    # set up solver
    if solver is None:
        solver = get_solver()
    optarg = solver.options

    # populate intitial properties throughout the system
    m.fs.feed.initialize(optarg=optarg)
    propagate_state(m.fs.s01)
    m.fs.separator.initialize(optarg=optarg, solver="ipopt-watertap")
    propagate_state(m.fs.s02)
    propagate_state(m.fs.s03)
    m.fs.EDstack.initialize(optarg=optarg)
    propagate_state(m.fs.s04)
    m.fs.product.initialize()
    propagate_state(m.fs.s05)
    m.fs.disposal.initialize()
    m.fs.costing.initialize()


def optimize_system(m, solver=None, checkpoint=None, fail_flag=True):

    # Below is an example of optimizing the operational voltage and cell pair number (which translates to membrane use)
    # Define a system with zero dof
    set_operating_conditions(m)

    # Set an objective function
    m.fs.objective = Objective(expr=m.fs.costing.LCOW)

    # Choose and unfix variables to be optimized
    m.fs.EDstack.voltage_applied[0].unfix()
    m.fs.EDstack.cell_pair_num.unfix()
    m.fs.EDstack.cell_pair_num.set_value(30)
    # Give narrower bounds to optimizing variables if available
    m.fs.EDstack.voltage_applied[0].setlb(0.5)
    m.fs.EDstack.voltage_applied[0].setub(20)
    m.fs.EDstack.cell_pair_num.setlb(5)
    m.fs.EDstack.cell_pair_num.setub(500)

    # Set a treatment goal
    # Example here is to reach a final product water containing NaCl = 1 g/L (from a 10 g/L feed)
    m.fs.product.properties[0].conc_mass_phase_comp["Liq", "Na_+"].fix(0.393)

    print("---report model statistics---\n ", report_statistics(m.fs))
    if solver is None:
        solver = get_solver()
    solve(m, solver=solver, tee=True)
    m.fs.EDstack.cell_pair_num.fix(round(value(m.fs.EDstack.cell_pair_num)))
    solve(m, solver=solver, tee=True)


def display_model_metrics(m):

    print("---Flow properties in feed, product and disposal---")

    fp = {
        "Feed": [
            value(m.fs.feed.properties[0].flow_vol_phase["Liq"]),
            value(
                sum(
                    m.fs.feed.properties[0].conc_mass_phase_comp["Liq", j]
                    for j in m.fs.properties.ion_set
                )
            ),
        ],
        "Product": [
            value(m.fs.product.properties[0].flow_vol_phase["Liq"]),
            value(m.fs.product_salinity),
        ],
        "Disposal": [
            value(m.fs.disposal.properties[0].flow_vol_phase["Liq"]),
            value(m.fs.disposal_salinity),
        ],
    }
    fp_table = DataFrame(
        data=fp,
        index=["Volume Flow Rate (m3/s)", "Total Ion Mass Concentration (kg/m3)"],
    )
    print(fp_table)

    print("---Performance Metrics---")

    pm_table = DataFrame(
        data=[
            value(m.fs.EDstack.recovery_mass_H2O[0]),
            value(m.fs.mem_area),
            value(m.fs.EDstack.cell_pair_num),
            value(m.fs.EDstack.voltage_applied[0]),
            value(m.fs.costing.specific_energy_consumption),
            value(m.fs.costing.LCOW),
        ],
        columns=["value"],
        index=[
            "Water recovery by mass",
            "Total membrane area (aem or cem), m2",
            "Cell pair number",
            "Operation Voltage, V",
            "Specific energy consumption, kWh/m3",
            "Levelized cost of water, $/m3",
        ],
    )
    print(pm_table)


if __name__ == "__main__":
    main()



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for CSTR unit model with injection.
Authors: Andrew Lee, Adam Atia, Vibhav Dabadghao
"""

import pytest
from pyomo.environ import (
    ConcreteModel,
    units,
    value,
    Objective,
)
from idaes.core import (
    FlowsheetBlock,
)

from watertap.unit_models.tests.unit_test_harness import UnitTestHarness
import idaes.core.util.scaling as iscale
from idaes.models.properties.examples.saponification_thermo import (
    SaponificationParameterBlock,
)
from idaes.models.properties.examples.saponification_reactions import (
    SaponificationReactionParameterBlock,
)
from idaes.core.util.exceptions import ConfigurationError
from watertap.core.solvers import get_solver

from watertap.unit_models.cstr_injection import CSTR_Injection, ElectricityConsumption
from idaes.core import UnitModelCostingBlock
from watertap.costing import WaterTAPCosting
from watertap.property_models.unit_specific.activated_sludge.asm1_properties import (
    ASM1ParameterBlock,
)
from watertap.property_models.unit_specific.activated_sludge.asm1_reactions import (
    ASM1ReactionParameterBlock,
)
from watertap.property_models.unit_specific.activated_sludge.asm2d_properties import (
    ASM2dParameterBlock,
)
from watertap.property_models.unit_specific.activated_sludge.asm2d_reactions import (
    ASM2dReactionParameterBlock,
)
from watertap.property_models.unit_specific.activated_sludge.modified_asm2d_properties import (
    ModifiedASM2dParameterBlock,
)
from watertap.property_models.unit_specific.activated_sludge.modified_asm2d_reactions import (
    ModifiedASM2dReactionParameterBlock,
)

from watertap.property_models.unit_specific.anaerobic_digestion.adm1_properties import (
    ADM1ParameterBlock,
)
from watertap.property_models.unit_specific.anaerobic_digestion.adm1_reactions import (
    ADM1ReactionParameterBlock,
)

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


# -----------------------------------------------------------------------------
def build():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = SaponificationParameterBlock()
    m.fs.reactions = SaponificationReactionParameterBlock(
        property_package=m.fs.properties
    )

    m.fs.unit = CSTR_Injection(
        property_package=m.fs.properties,
        reaction_package=m.fs.reactions,
        has_equilibrium_reactions=False,
        has_heat_transfer=True,
        has_heat_of_reaction=True,
        has_pressure_change=True,
    )

    m.fs.unit.inlet.flow_vol.fix(1.0e-03)
    m.fs.unit.inlet.conc_mol_comp[0, "H2O"].fix(55388.0)
    m.fs.unit.inlet.conc_mol_comp[0, "NaOH"].fix(100.0)
    m.fs.unit.inlet.conc_mol_comp[0, "EthylAcetate"].fix(100.0)
    m.fs.unit.inlet.conc_mol_comp[0, "SodiumAcetate"].fix(0.0)
    m.fs.unit.inlet.conc_mol_comp[0, "Ethanol"].fix(0.0)

    m.fs.unit.inlet.temperature.fix(303.15)
    m.fs.unit.inlet.pressure.fix(101325.0)

    m.fs.unit.volume.fix(1.5e-03)
    m.fs.unit.heat_duty.fix(0)
    m.fs.unit.deltaP.fix(0)
    m.fs.unit.injection.fix(0)

    iscale.set_scaling_factor(
        m.fs.unit.control_volume.properties_out[0.0].conc_mol_comp["H2O"], 1e-4
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.properties_out[0.0].pressure, 1e-5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.properties_out[0.0].flow_vol, 1e3
    )

    iscale.calculate_scaling_factors(m.fs.unit)

    return m


def build_ASM1():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = ASM1ParameterBlock()
    m.fs.reactions = ASM1ReactionParameterBlock(property_package=m.fs.properties)

    m.fs.unit = CSTR_Injection(
        property_package=m.fs.properties,
        reaction_package=m.fs.reactions,
        has_aeration=True,
        electricity_consumption=ElectricityConsumption.calculated,
    )

    m.fs.unit.inlet.flow_vol.fix(20648 * units.m**3 / units.day)
    m.fs.unit.inlet.temperature.fix(308.15 * units.K)
    m.fs.unit.inlet.pressure.fix(1 * units.atm)
    m.fs.unit.inlet.conc_mass_comp[0, "S_I"].fix(27 * units.g / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "S_S"].fix(58 * units.g / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "X_I"].fix(92 * units.g / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "X_S"].fix(363 * units.g / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "X_BH"].fix(50 * units.g / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "X_BA"].fix(0 * units.g / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "X_P"].fix(0 * units.g / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "S_O"].fix(0 * units.g / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "S_NO"].fix(0 * units.g / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "S_NH"].fix(23 * units.g / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "S_ND"].fix(5 * units.g / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "X_ND"].fix(16 * units.g / units.m**3)
    m.fs.unit.inlet.alkalinity.fix(7 * units.mol / units.m**3)

    m.fs.unit.volume.fix(500)
    m.fs.unit.injection.fix(0)
    m.fs.unit.injection[0, "Liq", "S_O"].fix(2e-3)

    # Set scaling factors for badly scaled variables
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.properties_out[0.0].pressure, 1e-5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.properties_out[0.0].conc_mass_comp["X_P"], 1e3
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "S_S"], 1e3
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "X_S"], 1e3
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "X_BH"], 1e3
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "X_P"], 1e3
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "S_O"], 1e3
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "S_NH"], 1e3
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "S_ND"], 1e3
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "X_ND"], 1e3
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "S_ALK"], 1e3
    )

    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_extent[0.0, "R4"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_extent[0.0, "R6"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_extent[0.0, "R7"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_extent[0.0, "R8"], 1e5
    )

    iscale.set_scaling_factor(
        m.fs.unit.control_volume.reactions[0.0].reaction_rate["R1"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.reactions[0.0].reaction_rate["R4"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.reactions[0.0].reaction_rate["R6"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.reactions[0.0].reaction_rate["R7"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.reactions[0.0].reaction_rate["R8"], 1e5
    )

    iscale.calculate_scaling_factors(m.fs.unit)

    return m


class TestCSTR_injection(UnitTestHarness):
    def configure(self):
        m = build()

        self.unit_solutions[m.fs.unit.outlet.pressure[0]] = 101325.0
        self.unit_solutions[m.fs.unit.outlet.temperature[0]] = 304.09
        self.unit_solutions[m.fs.unit.outlet.conc_mol_comp[0, "EthylAcetate"]] = 20.32
        self.unit_solutions[m.fs.unit.electricity_consumption[0]] = 0.03888
        self.unit_solutions[m.fs.unit.hydraulic_retention_time[0]] = 1.5
        self.unit_solutions[m.fs.unit.control_volume.heat_of_reaction[0]] = 3904.51

        # Conservation checks
        self.unit_solutions[m.fs.unit.outlet.flow_vol[0]] = value(
            m.fs.unit.inlet.flow_vol[0]
        )

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.inlet.flow_vol[0]
                * sum(
                    m.fs.unit.inlet.conc_mol_comp[0, j]
                    for j in m.fs.properties.component_list
                ),
                "out": m.fs.unit.outlet.flow_vol[0]
                * sum(
                    m.fs.unit.inlet.conc_mol_comp[0, j]
                    for j in m.fs.properties.component_list
                ),
            },
            "Check 2": {
                "in": (
                    m.fs.unit.inlet.flow_vol[0]
                    * m.fs.properties.dens_mol
                    * m.fs.properties.cp_mol
                    * (m.fs.unit.inlet.temperature[0] - m.fs.properties.temperature_ref)
                ),
                "out": m.fs.unit.outlet.flow_vol[0]
                * m.fs.properties.dens_mol
                * m.fs.properties.cp_mol
                * (m.fs.unit.outlet.temperature[0] - m.fs.properties.temperature_ref)
                - m.fs.unit.control_volume.heat_of_reaction[0],
            },
        }

        return m


class TestCosting_Saponification(UnitTestHarness):
    def configure(self):
        m = build()

        # Add unit model costing
        m.fs.costing = WaterTAPCosting()

        m.fs.unit.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
        m.fs.costing.cost_process()
        m.fs.costing.add_LCOW(m.fs.unit.control_volume.properties_out[0].flow_vol)
        m.objective = Objective(expr=m.fs.costing.LCOW)

        iscale.calculate_scaling_factors(m.fs.unit)

        iscale.set_scaling_factor(m.fs.unit.costing.capital_cost, 1e-2)

        m.fs.unit.initialize()

        solver.solve(m)

        self.unit_solutions[m.fs.unit.costing.capital_cost] = 7.75429 * 2
        self.unit_solutions[m.fs.costing.LCOW] = 0.00082698

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.inlet.flow_vol[0],
                "out": m.fs.unit.outlet.flow_vol[0],
            },
        }

        return m


class TestSaponification(object):
    @pytest.mark.unit
    def test_get_performance_contents(self):

        sapon = build()
        perf_dict = sapon.fs.unit._get_performance_contents()

        assert perf_dict == {
            "vars": {
                "Volume": sapon.fs.unit.volume[0],
                "Injection [('Liq', 'H2O')]": sapon.fs.unit.injection[0, "Liq", "H2O"],
                "Injection [('Liq', 'NaOH')]": sapon.fs.unit.injection[
                    0, "Liq", "NaOH"
                ],
                "Injection [('Liq', 'EthylAcetate')]": sapon.fs.unit.injection[
                    0, "Liq", "EthylAcetate"
                ],
                "Injection [('Liq', 'SodiumAcetate')]": sapon.fs.unit.injection[
                    0, "Liq", "SodiumAcetate"
                ],
                "Injection [('Liq', 'Ethanol')]": sapon.fs.unit.injection[
                    0, "Liq", "Ethanol"
                ],
                "Heat Duty": sapon.fs.unit.heat_duty[0],
                "Pressure Change": sapon.fs.unit.deltaP[0],
                "Electricity Consumption": sapon.fs.unit.electricity_consumption[0],
            }
        }


class TestCSTR_injection_ASM1(UnitTestHarness):
    def configure(self):
        m = build_ASM1()

        iscale.set_scaling_factor(
            m.fs.unit.control_volume.properties_out[0].conc_mass_comp["X_P"], 1e5
        )

        self.unit_solutions[m.fs.unit.outlet.pressure[0]] = 101325.0
        self.unit_solutions[m.fs.unit.outlet.temperature[0]] = 308.15
        self.unit_solutions[m.fs.unit.outlet.conc_mass_comp[0, "S_O"]] = 6.258e-3
        self.unit_solutions[m.fs.unit.electricity_consumption[0]] = 18.3765
        self.unit_solutions[m.fs.unit.hydraulic_retention_time[0]] = 2092.2123
        self.unit_solutions[m.fs.unit.KLa] = 8.2694

        # Conservation checks
        self.unit_solutions[m.fs.unit.outlet.flow_vol[0]] = value(
            m.fs.unit.inlet.flow_vol[0]
        )

        self.default_relative_tolerance = 1e-2

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.inlet.flow_vol[0]
                * sum(
                    m.fs.unit.inlet.conc_mass_comp[0, j]
                    for j in m.fs.properties.solute_set
                ),
                "out": m.fs.unit.outlet.flow_vol[0]
                * sum(
                    m.fs.unit.outlet.conc_mass_comp[0, j]
                    for j in m.fs.properties.solute_set
                )
                + sum(
                    m.fs.unit.control_volume.rate_reaction_generation[0, "Liq", j]
                    for j in m.fs.properties.solute_set
                ),
            },
        }

        return m


class TestCosting(UnitTestHarness):
    def configure(self):
        m = build_ASM1()

        # Add unit model costing
        m.fs.costing = WaterTAPCosting()

        m.fs.unit.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
        m.fs.costing.cost_process()
        m.fs.costing.add_LCOW(m.fs.unit.control_volume.properties_out[0].flow_vol)
        m.objective = Objective(expr=m.fs.costing.LCOW)

        iscale.set_scaling_factor(m.fs.unit.costing.capital_cost, 1e-7)
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.properties_out[0].conc_mass_comp["X_P"], 1e5
        )

        iscale.calculate_scaling_factors(m.fs.unit)

        m.fs.unit.initialize()

        solver.solve(m)

        self.unit_solutions[m.fs.unit.costing.capital_cost] = 613502.91662
        self.unit_solutions[m.fs.costing.LCOW] = 0.0132455

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.inlet.flow_vol[0],
                "out": m.fs.unit.outlet.flow_vol[0],
            },
        }

        return m


@pytest.mark.build
@pytest.mark.unit
def test_with_asm2d():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = ASM2dParameterBlock()
    m.fs.reactions = ASM2dReactionParameterBlock(property_package=m.fs.properties)

    m.fs.unit = CSTR_Injection(
        property_package=m.fs.properties,
        reaction_package=m.fs.reactions,
        has_aeration=True,
        electricity_consumption=ElectricityConsumption.calculated,
    )


@pytest.mark.build
@pytest.mark.unit
def test_with_mod_asm2d():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = ModifiedASM2dParameterBlock()
    m.fs.reactions = ModifiedASM2dReactionParameterBlock(
        property_package=m.fs.properties
    )

    m.fs.unit = CSTR_Injection(
        property_package=m.fs.properties,
        reaction_package=m.fs.reactions,
        has_aeration=True,
        electricity_consumption=ElectricityConsumption.calculated,
    )


@pytest.mark.unit
def test_error_without_oxygen():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = ADM1ParameterBlock()
    m.fs.reactions = ADM1ReactionParameterBlock(property_package=m.fs.properties)

    # Expect exception if has_aeration=True but S_O or S_O2 not listed in component_list of prop package.
    with pytest.raises(
        ConfigurationError,
        match="has_aeration was set to True, but the property package has neither 'S_O' nor 'S_O2' in its list of components.",
    ):
        m.fs.unit = CSTR_Injection(
            property_package=m.fs.properties,
            reaction_package=m.fs.reactions,
            has_aeration=True,
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Inherits from a modified CSTR model with injection terms. This model assumes oxygen will be injected.
"""

# Import Pyomo libraries
from pyomo.common.config import In

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
)


from watertap.unit_models.cstr_injection import (
    CSTR_InjectionData,
    ElectricityConsumption,
)

__author__ = "Adam Atia"


@declare_process_block_class("AerationTank")
class AerationTankData(CSTR_InjectionData):
    """
    CSTR Unit Model with Injection Class
    """

    CONFIG = CSTR_InjectionData.CONFIG()
    CONFIG.electricity_consumption = ElectricityConsumption.calculated
    CONFIG.get("has_aeration")._domain = In([True])
    CONFIG.has_aeration = True



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest
import abc

from pyomo.environ import Block, assert_optimal_termination, ComponentMap, value
from pyomo.util.check_units import assert_units_consistent
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
)
from watertap.core.solvers import get_solver
from idaes.core.util.testing import initialization_tester
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog


# -----------------------------------------------------------------------------
class UnitAttributeError(AttributeError):
    """
    WaterTAP exception for generic attribute errors arising from unit model testing.
    """


class UnitValueError(ValueError):
    """
    WaterTAP exception for generic value errors arising from unit model testing.
    """


class UnitRuntimeError(RuntimeError):
    """
    WaterTAP exception for generic runtime errors arising from unit model testing.
    """


class UnitTestHarness(abc.ABC):
    def configure_class(self):
        # string for solver, if None use WaterTAP default
        self.solver = None
        # dictionary for solver options, if None use WaterTAP default
        self.optarg = None

        # solution map from var to value
        self.unit_solutions = ComponentMap()

        # dictionary of expressions for conservation checks
        self.conservation_equality = {}

        # arguments for badly scaled variables
        self.default_large = 1e4
        self.default_small = 1e-3
        self.default_zero = 1e-10

        # arguments for solver tolerance
        self.default_absolute_tolerance = 1e-5
        self.default_relative_tolerance = 1e-3

        model = self.configure()
        if not hasattr(self, "unit_model_block"):
            self.unit_model_block = model.find_component("fs.unit")
            if self.unit_model_block is None:
                raise RuntimeError(
                    f"The {self.__class__.__name__}.configure method should either "
                    "set the attribute `unit_model_block` or name it `fs.unit`."
                )
        # keep the model so it does not get garbage collected
        self._model = model
        blk = self.unit_model_block

        # attaching objects to model to carry through in pytest frame
        # TODO: Consider removing these objects and directly calling self
        assert not hasattr(blk, "_test_objs")
        blk._test_objs = Block()
        blk._test_objs.solver = self.solver
        blk._test_objs.optarg = self.optarg
        blk._test_objs.unit_solutions = self.unit_solutions
        blk._test_objs.conservation_equality = self.conservation_equality

    @abc.abstractmethod
    def configure(self):
        """
        Placeholder method to allow user to setup test harness.

        The configure method must set the attributes:
        unit_solutions: ComponentMap of values for the specified variables

        The unit model tested should be named `fs.unit`, or this method
        should set the attribute `unit_model_block`.

        Returns:
            model: the top-level Pyomo model
        """

    @pytest.fixture(scope="class")
    def frame(self):
        self.configure_class()
        return self._model, self.unit_model_block

    @pytest.mark.unit
    def test_units_consistent(self, frame):
        m, unit_model = frame
        assert_units_consistent(unit_model)

    @pytest.mark.unit
    def test_dof(self, frame):
        m, unit_model = frame
        if degrees_of_freedom(unit_model) != 0:
            raise UnitAttributeError(
                "The unit has {dof} degrees of freedom when 0 is required."
                "".format(dof=degrees_of_freedom(unit_model))
            )

    @pytest.mark.component
    def test_initialization(self, frame):
        m, blk = frame
        initialization_tester(
            m,
            unit=blk,
            solver=blk._test_objs.solver,
            optarg=blk._test_objs.optarg,
            outlvl=idaeslog.DEBUG,
        )

    @pytest.mark.component
    def test_conservation(self, frame):
        self.configure_class()
        m, blk = frame

        conservation = blk._test_objs.conservation_equality

        if conservation == {}:
            raise NotImplementedError(
                "An expression must be provided for the inlet and outlet stream(s) in the conservation equality."
            )

        for key, expression in conservation.items():
            if "in" in expression and "out" in expression:
                inlet_expression = value(expression["in"])
                outlet_expression = value(expression["out"])
                try:
                    assert inlet_expression == pytest.approx(
                        outlet_expression,
                        abs=self.default_absolute_tolerance,
                        rel=self.default_relative_tolerance,
                    )
                except:
                    raise AssertionError(
                        f"In {key}, the inlet expression is equal to {inlet_expression}, "
                        f"but the outlet expression is equal to {outlet_expression}"
                    )
            else:
                raise AssertionError(
                    f"Ensure the name of the inlet expression in {key} is 'in' and the name of the "
                    f"outlet expression is 'out'."
                )

    @pytest.mark.component
    def test_unit_solutions(self, frame):
        self.configure_class()
        m, blk = frame
        solutions = blk._test_objs.unit_solutions

        # solve unit
        if blk._test_objs.solver is None:
            opt = get_solver()
        else:
            opt = get_solver(
                solver=blk._test_objs.solver, options=blk._test_objs.optarg
            )
        results = opt.solve(blk, tee=True)

        # check solve
        badly_scaled_vars = list(
            iscale.badly_scaled_var_generator(
                blk,
                large=self.default_large,
                small=self.default_small,
                zero=self.default_zero,
            )
        )
        if badly_scaled_vars:
            lines = [
                f"{x[0].name}\t{x[0].value}\tsf: {iscale.get_scaling_factor(x[0])}"
                for x in badly_scaled_vars
            ]
            msg = "One or more badly scaled variables found:\n"
            msg += "\n".join(lines)
            raise AssertionError(msg)

        assert_optimal_termination(results)

        # check results

        for var, val in solutions.items():
            comp_obj = None
            try:
                val = float(val)
            except:
                # expect the same API as pytest.approx
                comp_obj = val
                val = comp_obj.expected
            if comp_obj is None:
                comp_obj = pytest.approx(
                    val,
                    abs=self.default_absolute_tolerance,
                    rel=self.default_relative_tolerance,
                )
            if not comp_obj == value(var):
                raise AssertionError(f"{var}: Expected {val}, got {value(var)} instead")

    @pytest.mark.component
    def test_reporting(self, frame):
        m, blk = frame
        blk.report()



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Translator block representing the ADM1/ASM2d interface.
This is copied from the Generic template for a translator block.

Assumptions:
     * Steady-state only

Model formulated from:

Flores-Alsina, X., Solon, K., Mbamba, C.K., Tait, S., Gernaey, K.V., Jeppsson, U. and Batstone, D.J., 2016.
Modelling phosphorus (P), sulfur (S) and iron (Fe) interactions for dynamic simulations of anaerobic digestion processes.
Water Research, 95, pp.370-382.
"""

# Import Pyomo libraries
from pyomo.common.config import ConfigBlock, ConfigValue

# Import IDAES cores
from idaes.core import declare_process_block_class
from idaes.models.unit_models.translator import TranslatorData
from idaes.core.util.config import (
    is_reaction_parameter_block,
)
from idaes.core.util.model_statistics import degrees_of_freedom
from watertap.core.solvers import get_solver
import idaes.logger as idaeslog
import idaes.core.util.scaling as iscale

from idaes.core.util.exceptions import InitializationError

from pyomo.environ import (
    units as pyunits,
    check_optimal_termination,
    Set,
    Param,
)

__author__ = "Chenyu Wang, Marcus Holly, Xinhong Liu"


# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("Translator_ADM1_ASM2D")
class TranslatorDataADM1ASM2D(TranslatorData):
    """
    Translator block representing the ADM1/ASM2D interface
    """

    CONFIG = TranslatorData.CONFIG()

    CONFIG.declare(
        "inlet_reaction_package",
        ConfigValue(
            default=None,
            domain=is_reaction_parameter_block,
            description="Reaction package to use for control volume",
            doc="""Reaction parameter object used to define reaction calculations,
    **default** - None.
    **Valid values:** {
    **None** - no reaction package,
    **ReactionParameterBlock** - a ReactionParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "inlet_reaction_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing reaction packages",
            doc="""A ConfigBlock with arguments to be passed to a reaction block(s)
    and used when constructing these,
    **default** - None.
    **Valid values:** {
    see reaction package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "outlet_reaction_package",
        ConfigValue(
            default=None,
            domain=is_reaction_parameter_block,
            description="Reaction package to use for control volume",
            doc="""Reaction parameter object used to define reaction calculations,
        **default** - None.
        **Valid values:** {
        **None** - no reaction package,
        **ReactionParameterBlock** - a ReactionParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "outlet_reaction_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing reaction packages",
            doc="""A ConfigBlock with arguments to be passed to a reaction block(s)
        and used when constructing these,
        **default** - None.
        **Valid values:** {
        see reaction package for documentation.}""",
        ),
    )

    def build(self):
        """
        Begin building model.
        Args:
            None
        Returns:
            None
        """
        # Call UnitModel.build to setup dynamics
        super(TranslatorDataADM1ASM2D, self).build()

        eps = 0
        mw_p = 31 * pyunits.kg / pyunits.kmol
        mw_n = 14 * pyunits.kg / pyunits.kmol
        mw_c = 12 * pyunits.kg / pyunits.kmol
        mw_XPP = 300.41 * pyunits.kg / pyunits.kmol
        mw_k = 39.1 * pyunits.kg / pyunits.kmol
        mw_mg = 24.3 * pyunits.kg / pyunits.kmol

        @self.Constraint(
            self.flowsheet().time,
            doc="Equality volumetric flow equation",
        )
        def eq_flow_vol_rule(blk, t):
            return blk.properties_out[t].flow_vol == blk.properties_in[t].flow_vol

        @self.Constraint(
            self.flowsheet().time,
            doc="Equality temperature equation",
        )
        def eq_temperature_rule(blk, t):
            return blk.properties_out[t].temperature == blk.properties_in[t].temperature

        @self.Constraint(
            self.flowsheet().time,
            doc="Equality pressure equation",
        )
        def eq_pressure_rule(blk, t):
            return blk.properties_out[t].pressure == blk.properties_in[t].pressure

        # -------------------------------------------Step 1----------------------------------------------------------------#
        @self.Expression(
            self.flowsheet().time,
            doc="Biomass concentration (kgCOD/m3)",
        )
        def biomass(blk, t):
            return (
                blk.properties_in[t].conc_mass_comp["X_su"]
                + blk.properties_in[t].conc_mass_comp["X_aa"]
                + blk.properties_in[t].conc_mass_comp["X_fa"]
                + blk.properties_in[t].conc_mass_comp["X_c4"]
                + blk.properties_in[t].conc_mass_comp["X_pro"]
                + blk.properties_in[t].conc_mass_comp["X_ac"]
                + blk.properties_in[t].conc_mass_comp["X_h2"]
                + blk.properties_in[t].conc_mass_comp["X_PAO"]
            )

        @self.Expression(
            self.flowsheet().time, doc="S_ac concentration (kgCOD/m3) step 1"
        )
        def Sac_AD1(blk, t):
            return (
                blk.properties_in[t].conc_mass_comp["S_ac"]
                + blk.properties_in[t].conc_mass_comp["X_PHA"]
            )

        @self.Expression(
            self.flowsheet().time, doc="S_IC concentration at (kmolC/m3) step 1"
        )
        def SIC_AD1(blk, t):
            return (
                blk.config.inlet_reaction_package.Ci["X_su"] * blk.biomass[t]
                - (
                    blk.config.inlet_reaction_package.f_sI_xc
                    * blk.config.inlet_reaction_package.Ci["S_I"]
                    * blk.biomass[t]
                )
                - (
                    blk.config.inlet_reaction_package.f_ch_xc
                    * blk.biomass[t]
                    * blk.config.inlet_reaction_package.Ci["X_ch"]
                )
                - (
                    blk.config.inlet_reaction_package.f_pr_xc
                    * blk.biomass[t]
                    * blk.config.inlet_reaction_package.Ci["X_pr"]
                )
                - (
                    blk.config.inlet_reaction_package.f_li_xc
                    * blk.biomass[t]
                    * blk.config.inlet_reaction_package.Ci["X_li"]
                )
                - (
                    blk.config.inlet_reaction_package.f_xI_xc
                    * blk.biomass[t]
                    * blk.config.inlet_reaction_package.Ci["X_I"]
                )
                + (
                    blk.properties_in[t].conc_mass_comp["X_PHA"]
                    * blk.config.inlet_reaction_package.Ci["X_PHA"]
                )
                - (
                    blk.properties_in[t].conc_mass_comp["X_PHA"]
                    * blk.config.inlet_reaction_package.Ci["S_ac"]
                )
            )

        @self.Expression(
            self.flowsheet().time, doc="S_IN concentration (kmolN/m3) step 1"
        )
        def SIN_AD1(blk, t):
            return (
                blk.config.inlet_reaction_package.Ni["X_su"] * blk.biomass[t]
                - (
                    blk.config.inlet_reaction_package.f_sI_xc
                    * blk.biomass[t]
                    * blk.config.inlet_reaction_package.Ni["S_I"]
                )
                - (
                    blk.config.inlet_reaction_package.f_pr_xc
                    * blk.biomass[t]
                    * blk.config.inlet_reaction_package.Ni["X_pr"]
                )
                - (
                    blk.config.inlet_reaction_package.f_xI_xc
                    * blk.biomass[t]
                    * blk.config.inlet_reaction_package.Ni["X_I"]
                )
            )

        @self.Expression(
            self.flowsheet().time, doc="S_I concentration (kgCOD/m3) step 1"
        )
        def SI_AD1(blk, t):
            return blk.properties_in[t].conc_mass_comp["S_I"] + (
                blk.config.inlet_reaction_package.f_sI_xc * blk.biomass[t]
            )

        @self.Expression(
            self.flowsheet().time, doc="X_ch concentration (kgCOD/m3) step 1"
        )
        def Xch_AD1(blk, t):
            return blk.properties_in[t].conc_mass_comp["X_ch"] + (
                blk.config.inlet_reaction_package.f_ch_xc * blk.biomass[t]
            )

        @self.Expression(
            self.flowsheet().time, doc="X_pr concentration (kgCOD/m3) step 1"
        )
        def Xpr_AD1(blk, t):
            return blk.properties_in[t].conc_mass_comp["X_pr"] + (
                blk.config.inlet_reaction_package.f_pr_xc * blk.biomass[t]
            )

        @self.Expression(
            self.flowsheet().time, doc="X_li concentration (kgCOD/m3) step 1"
        )
        def Xli_AD1(blk, t):
            return blk.properties_in[t].conc_mass_comp["X_li"] + (
                blk.config.inlet_reaction_package.f_li_xc * blk.biomass[t]
            )

        @self.Expression(
            self.flowsheet().time, doc="X_I concentration (kgCOD/m3) step 1"
        )
        def XI_AD1(blk, t):
            return blk.properties_in[t].conc_mass_comp["X_I"] + (
                blk.config.inlet_reaction_package.f_xI_xc * blk.biomass[t]
            )

        @self.Expression(
            self.flowsheet().time, doc="S_IP concentration at (kmolP/m3) step 1"
        )
        def SIP_AD1(blk, t):
            return (
                blk.properties_in[t].conc_mass_comp["X_PP"] / mw_XPP
                + (blk.config.inlet_reaction_package.Pi["X_su"] * blk.biomass[t])
                - (
                    blk.config.inlet_reaction_package.f_sI_xc
                    * blk.biomass[t]
                    * blk.config.inlet_reaction_package.Pi["S_I"]
                )
                - (
                    blk.config.inlet_reaction_package.f_ch_xc
                    * blk.biomass[t]
                    * blk.config.inlet_reaction_package.P_ch
                )
                - (
                    blk.config.inlet_reaction_package.f_li_xc
                    * blk.biomass[t]
                    * blk.config.inlet_reaction_package.Pi["X_li"]
                )
                - (
                    blk.config.inlet_reaction_package.f_xI_xc
                    * blk.biomass[t]
                    * blk.config.inlet_reaction_package.Pi["X_I"]
                )
            )

        self.XPHA_AD1 = Param(
            initialize=eps,
            units=pyunits.kg / pyunits.m**3,
            mutable=True,
            doc="Mass concentration of X_PHA at step 1",
        )

        self.XPP_AD1 = Param(
            initialize=eps,
            units=pyunits.kmol / pyunits.m**3,
            mutable=True,
            doc="Molar concentration of X_PP at step 1",
        )

        self.XPAO_AD1 = Param(
            initialize=eps,
            units=pyunits.kg / pyunits.m**3,
            mutable=True,
            doc="Mass concentration of X_PAO at step 1",
        )

        # -------------------------------------------Step 2----------------------------------------------------------------#

        @self.Expression(
            self.flowsheet().time, doc="S_IC concentration at (kmolC/m3) step 2"
        )
        def SIC_AD2(blk, t):
            return (
                blk.Xch_AD1[t] * blk.config.inlet_reaction_package.Ci["X_ch"]
                + (blk.Xpr_AD1[t] * blk.config.inlet_reaction_package.Ci["X_pr"])
                + (blk.Xli_AD1[t] * blk.config.inlet_reaction_package.Ci["X_li"])
                - blk.config.outlet_reaction_package.i_CXS
                / mw_c
                * (blk.Xch_AD1[t] + blk.Xpr_AD1[t] + blk.Xli_AD1[t])
                + blk.properties_in[t].conc_mass_comp["S_su"]
                * blk.config.inlet_reaction_package.Ci["S_su"]
                + blk.properties_in[t].conc_mass_comp["S_aa"]
                * blk.config.inlet_reaction_package.Ci["S_aa"]
                + blk.properties_in[t].conc_mass_comp["S_fa"]
                * blk.config.inlet_reaction_package.Ci["S_fa"]
                - blk.config.outlet_reaction_package.i_CSF
                / mw_c
                * (
                    blk.properties_in[t].conc_mass_comp["S_su"]
                    + blk.properties_in[t].conc_mass_comp["S_aa"]
                    + blk.properties_in[t].conc_mass_comp["S_fa"]
                )
                + blk.properties_in[t].conc_mass_comp["S_va"]
                * blk.config.inlet_reaction_package.Ci["S_va"]
                + blk.properties_in[t].conc_mass_comp["S_bu"]
                * blk.config.inlet_reaction_package.Ci["S_bu"]
                + blk.properties_in[t].conc_mass_comp["S_pro"]
                * blk.config.inlet_reaction_package.Ci["S_pro"]
                + blk.Sac_AD1[t] * blk.config.inlet_reaction_package.Ci["S_ac"]
                - blk.config.outlet_reaction_package.i_CSA
                / mw_c
                * (
                    blk.properties_in[t].conc_mass_comp["S_va"]
                    + blk.properties_in[t].conc_mass_comp["S_bu"]
                    + blk.properties_in[t].conc_mass_comp["S_pro"]
                    + blk.Sac_AD1[t]
                )
            )

        @self.Expression(
            self.flowsheet().time, doc="S_IN concentration at (kmolN/m3) step 2"
        )
        def SIN_AD2(blk, t):
            return (
                blk.Xpr_AD1[t] * blk.config.inlet_reaction_package.Ni["X_pr"]
                - blk.config.outlet_reaction_package.i_NXS
                / mw_n
                * (blk.Xch_AD1[t] + blk.Xpr_AD1[t] + blk.Xli_AD1[t])
                + blk.properties_in[t].conc_mass_comp["S_aa"]
                * blk.config.inlet_reaction_package.Ni["S_aa"]
                - blk.config.outlet_reaction_package.i_NSF
                / mw_n
                * (
                    blk.properties_in[t].conc_mass_comp["S_su"]
                    + blk.properties_in[t].conc_mass_comp["S_fa"]
                    + blk.properties_in[t].conc_mass_comp["S_va"]
                )
            )

        @self.Expression(
            self.flowsheet().time, doc="S_IP concentration at (kmolP/m3) step 2"
        )
        def SIP_AD2(blk, t):
            return (
                self.XPP_AD1
                + blk.Xch_AD1[t] * blk.config.inlet_reaction_package.P_ch
                + blk.Xli_AD1[t] * blk.config.inlet_reaction_package.Pi["X_li"]
                - blk.config.outlet_reaction_package.i_PXS
                / mw_p
                * (blk.Xch_AD1[t] + blk.Xpr_AD1[t] + blk.Xli_AD1[t])
                - blk.config.outlet_reaction_package.i_PSF
                / mw_p
                * (
                    blk.properties_in[t].conc_mass_comp["S_su"]
                    + blk.properties_in[t].conc_mass_comp["S_fa"]
                    + blk.properties_in[t].conc_mass_comp["S_va"]
                )
            )

        @self.Expression(
            self.flowsheet().time, doc="S_K concentration at (kmolK/m3) step 2"
        )
        def SK_AD2(blk, t):
            return blk.properties_in[t].conc_mass_comp["S_K"] / mw_k + self.XPP_AD1 / 3

        @self.Expression(
            self.flowsheet().time, doc="S_Mg concentration at (kmolMg/m3) step 2"
        )
        def SMg_AD2(blk, t):
            return (
                blk.properties_in[t].conc_mass_comp["S_Mg"] / mw_mg + self.XPP_AD1 / 3
            )

        @self.Constraint(
            self.flowsheet().time, doc="S_F concentration output (kgCOD/m3)"
        )
        def SF_output(blk, t):
            return (
                blk.properties_out[t].conc_mass_comp["S_F"]
                == blk.properties_in[t].conc_mass_comp["S_su"]
                + blk.properties_in[t].conc_mass_comp["S_aa"]
                + blk.properties_in[t].conc_mass_comp["S_fa"]
            )

        @self.Constraint(
            self.flowsheet().time, doc="S_A concentration output (kgCOD/m3)"
        )
        def SA_output(blk, t):
            return (
                blk.properties_out[t].conc_mass_comp["S_A"]
                == blk.properties_in[t].conc_mass_comp["S_va"]
                + blk.properties_in[t].conc_mass_comp["S_bu"]
                + blk.properties_in[t].conc_mass_comp["S_pro"]
                + blk.Sac_AD1[t]
            )

        @self.Constraint(
            self.flowsheet().time, doc="S_I concentration output (kgCOD/m3)"
        )
        def SI_output(blk, t):
            return blk.properties_out[t].conc_mass_comp["S_I"] == blk.SI_AD1[t]

        @self.Constraint(
            self.flowsheet().time, doc="S_NH4 concentration output (kgN/m3)"
        )
        def SNH4_output(blk, t):
            return blk.properties_out[t].conc_mass_comp["S_NH4"] == mw_n * (
                blk.properties_in[t].conc_mass_comp["S_IN"] / mw_n
                + blk.SIN_AD1[t]
                + blk.SIN_AD2[t]
            )

        @self.Constraint(
            self.flowsheet().time, doc="S_PO4 concentration output (kgP/m3)"
        )
        def SPO4_output(blk, t):
            return blk.properties_out[t].conc_mass_comp["S_PO4"] == mw_p * (
                blk.properties_in[t].conc_mass_comp["S_IP"] / mw_p
                + blk.SIP_AD1[t]
                + blk.SIP_AD2[t]
            )

        @self.Constraint(
            self.flowsheet().time, doc="S_IC concentration output (kgC/m3)"
        )
        def SIC_output(blk, t):
            return blk.properties_out[t].conc_mass_comp["S_IC"] == mw_c * (
                blk.properties_in[t].conc_mass_comp["S_IC"] / mw_c
                + blk.SIC_AD1[t]
                + blk.SIC_AD2[t]
            )

        @self.Constraint(
            self.flowsheet().time, doc="X_I concentration output (kgCOD/m3)"
        )
        def XI_output(blk, t):
            return blk.properties_out[t].conc_mass_comp["X_I"] == blk.XI_AD1[t]

        @self.Constraint(
            self.flowsheet().time, doc="X_S concentration output (kgCOD/m3)"
        )
        def XS_output(blk, t):
            return (
                blk.properties_out[t].conc_mass_comp["X_S"]
                == blk.Xch_AD1[t] + blk.Xpr_AD1[t] + blk.Xli_AD1[t]
            )

        @self.Constraint(
            self.flowsheet().time, doc="S_K concentration output (kgCOD/m3)"
        )
        # TODO: Need to add precipitation term for X_Kstruv
        def SK_output(blk, t):
            return blk.properties_out[t].conc_mass_comp["S_K"] == blk.SK_AD2[t] * mw_k

        @self.Constraint(
            self.flowsheet().time, doc="S_Mg concentration output (kgCOD/m3)"
        )
        def SMg_output(blk, t):
            return (
                blk.properties_out[t].conc_mass_comp["S_Mg"] == blk.SMg_AD2[t] * mw_mg
            )

        self.zero_flow_components = Set(
            initialize=[
                "S_O2",
                "S_N2",
                "S_NO3",
                "X_H",
                "X_PAO",
                "X_PP",
                "X_PHA",
                "X_AUT",
            ]
        )

        @self.Constraint(
            self.flowsheet().time,
            self.zero_flow_components,
            doc="Components with no flow equation",
        )
        def return_zero_flow_comp(blk, t, i):
            return (
                blk.properties_out[t].conc_mass_comp[i]
                == 1e-10 * pyunits.kg / pyunits.m**3
            )

        iscale.set_scaling_factor(self.properties_out[0].flow_vol, 1e5)

    def initialize_build(
        self,
        state_args_in=None,
        state_args_out=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        This method calls the initialization method of the state blocks.

        Keyword Arguments:
            state_args_in : a dict of arguments to be passed to the inlet
                property package (to provide an initial state for
                initialization (see documentation of the specific
                property package) (default = None).
            state_args_out : a dict of arguments to be passed to the outlet
                property package (to provide an initial state for
                initialization (see documentation of the specific
                property package) (default = None).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)

        Returns:
            None
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # Initialize state block
        flags = self.properties_in.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args_in,
            hold_state=True,
        )

        self.properties_out.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args_out,
        )

        if degrees_of_freedom(self) != 0:
            raise Exception(
                f"{self.name} degrees of freedom were not 0 at the beginning "
                f"of initialization. DoF = {degrees_of_freedom(self)}"
            )

        with idaeslog.solver_log(init_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)

        self.properties_in.release_state(flags=flags, outlvl=outlvl)

        init_log.info(f"Initialization Complete: {idaeslog.condition(res)}")

        if not check_optimal_termination(res):
            raise InitializationError(
                f"{self.name} failed to initialize successfully. Please check "
                f"the output logs for more information."
            )



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for general zero-order costing methods
"""
import pytest

from pyomo.environ import (
    Block,
    check_optimal_termination,
    ConcreteModel,
    Constraint,
    Expression,
    Set,
    units as pyunits,
    value,
    Var,
)
from pyomo.util.check_units import assert_units_consistent
from pyomo.common.config import ConfigValue

from idaes.core import FlowsheetBlock, declare_process_block_class
from watertap.core.solvers import get_solver
from idaes.core import UnitModelCostingBlock
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_unfixed_variables,
)
from idaes.core.util.misc import add_object_reference

from watertap.costing.zero_order_costing import (
    ZeroOrderCosting,
    _load_case_study_definition,
)
from watertap.core.zero_order_base import ZeroOrderBaseData
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.core.wt_database import Database
from watertap.unit_models.zero_order import ChemicalAdditionZO, NanofiltrationZO


@declare_process_block_class("DerivedZOBase")
class DerivedZOBaseData(ZeroOrderBaseData):
    def build(self):
        super().build()


solver = get_solver()


# TODO: Test general costing methods


class TestGeneralMethods:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.params = WaterParameterBlock(solute_list=["A", "B", "C"])

        m.fs.frame = ZeroOrderCosting()

        # Dummy a unit model to use with _get_tech_parameters
        m.fs.dummy_unit = DerivedZOBase(property_package=m.fs.params)
        m.fs.dummy_unit._tech_type = "test_tech"
        m.fs.dummy_unit.config.declare("flowsheet_costing_block", ConfigValue())
        m.fs.dummy_unit.config.flowsheet_costing_block = m.fs.frame
        add_object_reference(m.fs.dummy_unit, "unit_model", m.fs.dummy_unit)

        return m

    @pytest.mark.unit
    def test_load_case_study_definition(self, model):
        assert model.fs.frame.config.case_study_definition is None

        def_dict = _load_case_study_definition(model.fs.frame)

        assert isinstance(def_dict, dict)

        model.fs.frame.config.case_study_definition = "foo"
        with pytest.raises(
            OSError,
            match="Could not find specified case study "
            "definition file. Please check the path provided.",
        ):
            _load_case_study_definition(model.fs.frame)

    @pytest.mark.unit
    def test_build_global_params(self, model):
        assert model.fs.frame.base_currency == pyunits.MUSD_2018
        assert model.fs.frame.base_period == pyunits.year

        assert len(model.fs.frame.defined_flows) == 21
        for f in model.fs.frame.defined_flows:
            assert f in [
                "heat",
                "electricity",
                "acetic_acid",
                "activated_carbon",
                "alum",
                "ammonia",
                "anthracite",
                "anti-scalant",
                "cationic_polymer",
                "caustic_soda",
                "chlorine",
                "ferric_chloride",
                "hydrochloric_acid",
                "hydrogen_peroxide",
                "ion_exchange_resin",
                "lime",
                "phosphoric_acid",
                "sand",
                "sodium_bisulfite",
                "sodium_chloride",
                "sulfuric_acid",
            ]

        # factor capital annualization
        assert number_unfixed_variables(model.fs.frame) == 1

        assert isinstance(model.fs.frame.plant_lifetime, Var)
        assert value(model.fs.frame.plant_lifetime) == 30
        assert isinstance(model.fs.frame.utilization_factor, Var)
        assert value(model.fs.frame.utilization_factor) == 1

        assert isinstance(model.fs.frame.land_cost_percent_FCI, Var)
        assert value(model.fs.frame.land_cost_percent_FCI) == 0.0015
        assert isinstance(model.fs.frame.working_capital_percent_FCI, Var)
        assert value(model.fs.frame.working_capital_percent_FCI) == 0.05
        assert isinstance(model.fs.frame.salaries_percent_FCI, Var)
        assert value(model.fs.frame.salaries_percent_FCI) == 0.001
        assert isinstance(model.fs.frame.benefit_percent_of_salary, Var)
        assert value(model.fs.frame.benefit_percent_of_salary) == 0.9
        assert isinstance(model.fs.frame.maintenance_costs_percent_FCI, Var)
        assert value(model.fs.frame.maintenance_costs_percent_FCI) == 0.008
        assert isinstance(model.fs.frame.laboratory_fees_percent_FCI, Var)
        assert value(model.fs.frame.laboratory_fees_percent_FCI) == 0.003
        assert isinstance(model.fs.frame.insurance_and_taxes_percent_FCI, Var)
        assert value(model.fs.frame.insurance_and_taxes_percent_FCI) == 0.002

        assert isinstance(model.fs.frame.wacc, Var)
        assert value(model.fs.frame.wacc) == 0.05
        assert isinstance(model.fs.frame.capital_recovery_factor, Var)
        assert value(model.fs.frame.capital_recovery_factor) == pytest.approx(
            0.0650514, rel=1e-5
        )

        assert isinstance(model.fs.frame.TPEC, Var)
        assert value(model.fs.frame.TPEC) == 3.4
        assert isinstance(model.fs.frame.TIC, Var)
        assert value(model.fs.frame.TIC) == 1.65

    @pytest.mark.unit
    def test_get_tech_parameters_first_call(self, model):
        assert not hasattr(model.fs.frame, "test_tech")

        parameters = {
            "capital_cost": {
                "capital_a_parameter": {"value": 1, "units": "m"},
                "capital_b_parameter": {"value": 7, "units": "K"},
                "reference_state": {"value": 22, "units": "mol"},
            }
        }

        A, B, R = model.fs.dummy_unit._get_tech_parameters(
            model.fs.dummy_unit,
            parameters,
            None,
            ["capital_a_parameter", "capital_b_parameter", "reference_state"],
        )

        assert isinstance(model.fs.frame.test_tech, Block)
        assert isinstance(model.fs.frame.test_tech.subtype_set, Set)
        assert len(model.fs.frame.test_tech.subtype_set) == 1
        assert None in model.fs.frame.test_tech.subtype_set

        assert isinstance(model.fs.frame.test_tech.capital_a_parameter, Var)
        assert (
            model.fs.frame.test_tech.capital_a_parameter.index_set()
            is model.fs.frame.test_tech.subtype_set
        )
        assert value(model.fs.frame.test_tech.capital_a_parameter[None]) == 1
        assert model.fs.frame.test_tech.capital_a_parameter[None].fixed

        assert isinstance(model.fs.frame.test_tech.capital_b_parameter, Var)
        assert (
            model.fs.frame.test_tech.capital_b_parameter.index_set()
            is model.fs.frame.test_tech.subtype_set
        )
        assert value(model.fs.frame.test_tech.capital_b_parameter[None]) == 7
        assert model.fs.frame.test_tech.capital_b_parameter[None].fixed

        assert isinstance(model.fs.frame.test_tech.reference_state, Var)
        assert (
            model.fs.frame.test_tech.reference_state.index_set()
            is model.fs.frame.test_tech.subtype_set
        )
        assert value(model.fs.frame.test_tech.reference_state[None]) == 22
        assert model.fs.frame.test_tech.reference_state[None].fixed

        assert A is model.fs.frame.test_tech.capital_a_parameter[None]
        assert B is model.fs.frame.test_tech.capital_b_parameter[None]
        assert R is model.fs.frame.test_tech.reference_state[None]

    @pytest.mark.unit
    def test_get_tech_parameters_second_call(self, model):
        # Use different parameter values- these should be ignored
        parameters = {
            "capital_cost": {
                "capital_a_parameter": {"value": 100, "units": "m"},
                "capital_b_parameter": {"value": 700, "units": "K"},
                "reference_state": {"value": 2200, "units": "mol"},
            }
        }

        A, B, R = model.fs.dummy_unit._get_tech_parameters(
            model.fs.dummy_unit,
            parameters,
            None,
            ["capital_a_parameter", "capital_b_parameter", "reference_state"],
        )

        assert isinstance(model.fs.frame.test_tech, Block)
        assert isinstance(model.fs.frame.test_tech.subtype_set, Set)
        assert len(model.fs.frame.test_tech.subtype_set) == 1
        assert None in model.fs.frame.test_tech.subtype_set

        assert isinstance(model.fs.frame.test_tech.capital_a_parameter, Var)
        assert (
            model.fs.frame.test_tech.capital_a_parameter.index_set()
            is model.fs.frame.test_tech.subtype_set
        )
        assert value(model.fs.frame.test_tech.capital_a_parameter[None]) == 1
        assert model.fs.frame.test_tech.capital_a_parameter[None].fixed

        assert isinstance(model.fs.frame.test_tech.capital_b_parameter, Var)
        assert (
            model.fs.frame.test_tech.capital_b_parameter.index_set()
            is model.fs.frame.test_tech.subtype_set
        )
        assert value(model.fs.frame.test_tech.capital_b_parameter[None]) == 7
        assert model.fs.frame.test_tech.capital_b_parameter[None].fixed

        assert isinstance(model.fs.frame.test_tech.reference_state, Var)
        assert (
            model.fs.frame.test_tech.reference_state.index_set()
            is model.fs.frame.test_tech.subtype_set
        )
        assert value(model.fs.frame.test_tech.reference_state[None]) == 22
        assert model.fs.frame.test_tech.reference_state[None].fixed

        assert A is model.fs.frame.test_tech.capital_a_parameter[None]
        assert B is model.fs.frame.test_tech.capital_b_parameter[None]
        assert R is model.fs.frame.test_tech.reference_state[None]

    @pytest.mark.unit
    def test_get_tech_parameters_second_subtype(self, model):
        # Different parameters should be used this time
        parameters = {
            "capital_cost": {
                "capital_a_parameter": {"value": 100, "units": "m"},
                "capital_b_parameter": {"value": 700, "units": "K"},
                "reference_state": {"value": 2200, "units": "mol"},
            }
        }

        # Provide subtype foo
        A, B, R = model.fs.dummy_unit._get_tech_parameters(
            model.fs.dummy_unit,
            parameters,
            "foo",
            ["capital_a_parameter", "capital_b_parameter", "reference_state"],
        )

        assert isinstance(model.fs.frame.test_tech, Block)
        assert isinstance(model.fs.frame.test_tech.subtype_set, Set)
        assert len(model.fs.frame.test_tech.subtype_set) == 2
        assert "foo" in model.fs.frame.test_tech.subtype_set

        assert isinstance(model.fs.frame.test_tech.capital_a_parameter, Var)
        assert (
            model.fs.frame.test_tech.capital_a_parameter.index_set()
            is model.fs.frame.test_tech.subtype_set
        )
        assert value(model.fs.frame.test_tech.capital_a_parameter["foo"]) == 100
        assert model.fs.frame.test_tech.capital_a_parameter["foo"].fixed

        assert isinstance(model.fs.frame.test_tech.capital_b_parameter, Var)
        assert (
            model.fs.frame.test_tech.capital_b_parameter.index_set()
            is model.fs.frame.test_tech.subtype_set
        )
        assert value(model.fs.frame.test_tech.capital_b_parameter["foo"]) == 700
        assert model.fs.frame.test_tech.capital_b_parameter["foo"].fixed

        assert isinstance(model.fs.frame.test_tech.reference_state, Var)
        assert (
            model.fs.frame.test_tech.reference_state.index_set()
            is model.fs.frame.test_tech.subtype_set
        )
        assert value(model.fs.frame.test_tech.reference_state["foo"]) == 2200
        assert model.fs.frame.test_tech.reference_state["foo"].fixed

        assert A is model.fs.frame.test_tech.capital_a_parameter["foo"]
        assert B is model.fs.frame.test_tech.capital_b_parameter["foo"]
        assert R is model.fs.frame.test_tech.reference_state["foo"]


class TestWorkflow:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.params = WaterParameterBlock(solute_list=["sulfur", "toc", "tss"])

        m.fs.costing = ZeroOrderCosting()

        return m

    @pytest.mark.component
    def test_nf_costing(self, model):
        model.fs.unit1 = NanofiltrationZO(
            property_package=model.fs.params, database=model.db
        )

        model.fs.unit1.inlet.flow_mass_comp[0, "H2O"].fix(10000)
        model.fs.unit1.inlet.flow_mass_comp[0, "sulfur"].fix(1)
        model.fs.unit1.inlet.flow_mass_comp[0, "toc"].fix(2)
        model.fs.unit1.inlet.flow_mass_comp[0, "tss"].fix(3)
        model.fs.unit1.load_parameters_from_database()
        assert degrees_of_freedom(model.fs.unit1) == 0

        model.fs.unit1.costing = UnitModelCostingBlock(
            flowsheet_costing_block=model.fs.costing
        )

        assert isinstance(model.fs.costing.nanofiltration, Block)
        assert isinstance(model.fs.costing.nanofiltration.capital_a_parameter, Var)
        assert isinstance(model.fs.costing.nanofiltration.capital_b_parameter, Var)
        assert isinstance(model.fs.costing.nanofiltration.reference_state, Var)

        assert isinstance(model.fs.unit1.costing.capital_cost, Var)
        assert isinstance(model.fs.unit1.costing.capital_cost_constraint, Constraint)

        assert_units_consistent(model.fs)
        assert degrees_of_freedom(model.fs.unit1) == 0

        assert (
            model.fs.unit1.electricity[0]
            in model.fs.costing._registered_flows["electricity"]
        )

    @pytest.mark.component
    def test_chem_addition_costing(self, model):
        model.fs.unit2 = ChemicalAdditionZO(
            property_package=model.fs.params, process_subtype="alum", database=model.db
        )

        model.fs.unit2.inlet.flow_mass_comp[0, "H2O"].fix(1000)
        model.fs.unit2.inlet.flow_mass_comp[0, "sulfur"].fix(1)
        model.fs.unit2.inlet.flow_mass_comp[0, "toc"].fix(2)
        model.fs.unit2.inlet.flow_mass_comp[0, "tss"].fix(3)
        model.fs.unit2.load_parameters_from_database()
        assert degrees_of_freedom(model.fs.unit2) == 0

        model.fs.unit2.costing = UnitModelCostingBlock(
            flowsheet_costing_block=model.fs.costing,
            costing_method_arguments={"number_of_parallel_units": 2},
        )

        assert isinstance(model.fs.costing.chemical_addition, Block)
        assert isinstance(model.fs.costing.chemical_addition.capital_a_parameter, Var)
        assert isinstance(model.fs.costing.chemical_addition.capital_b_parameter, Var)

        assert isinstance(model.fs.unit2.costing.capital_cost, Var)
        assert isinstance(model.fs.unit2.costing.capital_cost_constraint, Constraint)

        assert_units_consistent(model.fs)
        assert degrees_of_freedom(model.fs.unit2) == 0

        assert (
            model.fs.unit2.electricity[0]
            is model.fs.costing._registered_flows["electricity"][1]
        )
        assert pytest.approx(1.006e3 * 10 / 0.5, rel=1e-8) == value(
            pyunits.convert(
                model.fs.costing._registered_flows["alum"][0],
                to_units=pyunits.mg / pyunits.s,
            )
        )

    @pytest.mark.component
    def test_process_costing(self, model):
        model.fs.costing.cost_process()

        assert isinstance(model.fs.costing.aggregate_capital_cost, Var)
        assert isinstance(model.fs.costing.aggregate_fixed_operating_cost, Var)
        assert isinstance(model.fs.costing.aggregate_variable_operating_cost, Var)
        assert isinstance(model.fs.costing.aggregate_flow_electricity, Var)
        assert isinstance(model.fs.costing.aggregate_flow_costs, Var)

        assert isinstance(model.fs.costing.land_cost, Expression)
        assert isinstance(model.fs.costing.working_capital, Expression)
        assert isinstance(model.fs.costing.total_capital_cost, Var)

        assert isinstance(model.fs.costing.total_capital_cost_constraint, Constraint)

        assert_units_consistent(model.fs)
        assert degrees_of_freedom(model.fs) == 0

    @pytest.mark.component
    def test_add_LCOW(self, model):
        model.fs.costing.add_LCOW(model.fs.unit1.properties_in[0].flow_vol)

        assert isinstance(model.fs.costing.LCOW, Expression)

        assert_units_consistent(model.fs)
        assert degrees_of_freedom(model.fs) == 0

    @pytest.mark.component
    def test_add_electricity_intensity(self, model):
        model.fs.costing.add_electricity_intensity(
            model.fs.unit1.properties_in[0].flow_vol
        )

        assert isinstance(model.fs.costing.electricity_intensity, Expression)

        assert_units_consistent(model.fs)
        assert degrees_of_freedom(model.fs) == 0

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, model):
        model.fs.unit1.initialize()
        model.fs.unit2.initialize()

        model.fs.costing.initialize()

        assert pytest.approx(0, abs=1e-5) == value(
            model.fs.costing.total_capital_cost_constraint
        )

        assert pytest.approx(0, abs=1e-5) == value(
            model.fs.costing.total_operating_cost_constraint
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model, tee=True)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        # Note all dollar values are in millions of dollars
        assert pytest.approx(630.596, rel=1e-5) == value(
            model.fs.costing.total_capital_cost
        )

        assert pytest.approx(8333.42, rel=1e-5) == value(
            model.fs.costing.aggregate_flow_electricity
        )
        assert pytest.approx(20.12, rel=1e-5) == value(
            model.fs.costing.aggregate_flow_alum
        )

        # Note units (M$)
        assert pytest.approx(1.73278e-7, rel=1e-5) == value(model.fs.costing.LCOW)

        assert pytest.approx(0.231345, rel=1e-5) == value(
            model.fs.costing.electricity_intensity
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pyomo.environ as pyo
from ..util import (
    register_costing_parameter_block,
    make_capital_cost_var,
)


def build_cstr_cost_param_block(blk):
    # Source: C.-C. TANG, Mathematical Models and Optimization Techniques for Use in Analysis and Design of Wastewater Treatment Systems, Ph.D., University of Illinois at Urbana-Champaign, 1984. https://www.proquest.com/docview/303308678/abstract/1DA7388DED324E60PQ/1 (accessed December 14, 2023).

    blk.capital_a_parameter = pyo.Var(
        initialize=1246.1,
        doc="A parameter for capital cost, converted from 1971 (from 461) to 1990",
        units=pyo.units.USD_1990,
    )
    blk.capital_b_parameter = pyo.Var(
        initialize=0.71,
        doc="B parameter for capital cost",
        units=pyo.units.dimensionless,
    )


@register_costing_parameter_block(
    build_rule=build_cstr_cost_param_block,
    parameter_block_name="cstr",
)
def cost_cstr(blk):
    """
    CSTR costing method
    """

    make_capital_cost_var(blk)
    blk.costing_package.add_cost_factor(blk, "TIC")
    cost_blk = blk.costing_package.cstr

    blk.capital_cost_constraint = pyo.Constraint(
        expr=blk.capital_cost
        == blk.cost_factor
        * pyo.units.convert(
            cost_blk.capital_a_parameter
            * (blk.unit_model.control_volume.volume[0] / pyo.units.m**3)
            ** cost_blk.capital_b_parameter,
            to_units=blk.costing_package.base_currency,
        )
    )



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains a zero-order representation of a surface discharge unit
operation.
"""

import pyomo.environ as pyo
from pyomo.environ import Reference, units as pyunits, Var
from idaes.core import declare_process_block_class

from watertap.core import build_pt, pump_electricity, ZeroOrderBaseData

# Some more information about this module
__author__ = "Travis Arnold"


@declare_process_block_class("SurfaceDischargeZO")
class SurfaceDischargeData(ZeroOrderBaseData):
    """
    Zero-Order model for a surface discharge unit operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "surface_discharge"
        build_pt(self)
        self._Q = Reference(self.properties[:].flow_vol)

        pump_electricity(self, self._Q)

        self.pipe_distance = Var(
            self.flowsheet().config.time, units=pyunits.miles, doc="Piping distance"
        )

        self.pipe_diameter = Var(
            self.flowsheet().config.time, units=pyunits.inches, doc="Pipe diameter"
        )

        self._fixed_perf_vars.append(self.pipe_distance)
        self._fixed_perf_vars.append(self.pipe_diameter)

        self._perf_var_dict["Pipe Distance"] = self.pipe_distance
        self._perf_var_dict["Pipe Diameter"] = self.pipe_diameter

    @property
    def default_costing_method(self):
        return self.cost_surface_discharge

    @staticmethod
    def cost_surface_discharge(blk):
        """
        General method for costing surface discharge. Capital cost is based on
        construction and pipe costs.
        """

        t0 = blk.flowsheet().time.first()

        # Get parameter dict from database
        parameter_dict = blk.unit_model.config.database.get_unit_operation_parameters(
            blk.unit_model._tech_type, subtype=blk.unit_model.config.process_subtype
        )

        # Get costing parameter sub-block for this technology
        A, B, pipe_cost_basis, ref_state = blk.unit_model._get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            [
                "capital_a_parameter",
                "capital_b_parameter",
                "pipe_cost_basis",
                "reference_state",
            ],
        )

        # Add cost variable and constraint
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation",
        )

        expr = pyo.units.convert(
            A
            * pyo.units.convert(
                blk.unit_model.properties[t0].flow_vol / ref_state,
                to_units=pyo.units.dimensionless,
            )
            ** B,
            to_units=blk.config.flowsheet_costing_block.base_currency,
        ) + pyo.units.convert(
            pipe_cost_basis
            * blk.unit_model.pipe_distance[t0]
            * blk.unit_model.pipe_diameter[t0],
            to_units=blk.config.flowsheet_costing_block.base_currency,
        )

        blk.costing_package.add_cost_factor(
            blk, parameter_dict["capital_cost"]["cost_factor"]
        )

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == blk.cost_factor * expr
        )

        # Register flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity"
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
from pyomo.environ import ConcreteModel

from watertap.core.solvers import get_solver

from idaes.core import FlowsheetBlock

import idaes.core.util.scaling as iscale

from watertap.unit_models.reverse_osmosis_base import TransportModel, ModuleType

import watertap.property_models.NaCl_prop_pack as props

from watertap.unit_models.tests.unit_test_harness import UnitTestHarness

from watertap.unit_models.reverse_osmosis_1D import (
    ReverseOsmosis1D,
    ConcentrationPolarizationType,
    MassTransferCoefficient,
    PressureChangeType,
)

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()
# -----------------------------------------------------------------------------


def build():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = props.NaClParameterBlock()

    m.fs.unit = ReverseOsmosis1D(
        property_package=m.fs.properties,
        has_pressure_change=True,
        concentration_polarization_type=ConcentrationPolarizationType.calculated,
        mass_transfer_coefficient=MassTransferCoefficient.calculated,
        pressure_change_type=PressureChangeType.calculated,
        finite_elements=3,
        has_full_reporting=True,
    )

    # fully specify system
    feed_flow_mass = 1000 / 3600
    feed_mass_frac_NaCl = 0.034283
    feed_pressure = 70e5

    feed_temperature = 273.15 + 25
    A = 4.2e-12
    B = 3.5e-8
    pressure_atmospheric = 1e5
    feed_mass_frac_H2O = 1 - feed_mass_frac_NaCl

    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "NaCl"].fix(
        feed_flow_mass * feed_mass_frac_NaCl
    )

    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
        feed_flow_mass * feed_mass_frac_H2O
    )

    m.fs.unit.inlet.pressure[0].fix(feed_pressure)
    m.fs.unit.inlet.temperature[0].fix(feed_temperature)
    m.fs.unit.A_comp.fix(A)
    m.fs.unit.B_comp.fix(B)
    m.fs.unit.permeate.pressure[0].fix(pressure_atmospheric)
    m.fs.unit.feed_side.N_Re[0, 0].fix(400)
    m.fs.unit.recovery_mass_phase_comp[0, "Liq", "H2O"].fix(0.5)
    m.fs.unit.feed_side.spacer_porosity.fix(0.97)
    m.fs.unit.feed_side.channel_height.fix(0.001)

    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e1, index=("Liq", "H2O")
    )
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e3, index=("Liq", "NaCl")
    )

    iscale.calculate_scaling_factors(m.fs.unit)

    return m


class TestReverseOsmosis1D(UnitTestHarness):
    def configure(self):
        m = build()
        x_interface_in = m.fs.unit.feed_side.length_domain.at(2)

        self.unit_solutions[m.fs.unit.deltaP[0]] = -1.75535e5
        self.unit_solutions[
            m.fs.unit.flux_mass_phase_comp[0, x_interface_in, "Liq", "H2O"]
        ] = 0.00799178
        self.unit_solutions[
            m.fs.unit.flux_mass_phase_comp[0, x_interface_in, "Liq", "NaCl"]
        ] = 2.113811e-6
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp[0, 1, "Liq", "H2O"]] = (
            0.00248645
        )
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp[0, 1, "Liq", "NaCl"]] = (
            2.593074e-6
        )
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp_avg[0, "Liq", "H2O"]] = (
            0.005036466
        )
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp_avg[0, "Liq", "NaCl"]] = (
            2.372961e-6
        )
        self.unit_solutions[
            m.fs.unit.mixed_permeate[0].flow_mass_phase_comp["Liq", "H2O"]
        ] = 0.1341274
        self.unit_solutions[
            m.fs.unit.mixed_permeate[0].flow_mass_phase_comp["Liq", "NaCl"]
        ] = 6.319489e-5
        self.unit_solutions[m.fs.unit.feed_side.N_Re_avg[0]] = 371.01255
        self.unit_solutions[m.fs.unit.feed_side.K_avg[0, "NaCl"]] = 2.985544e-5
        self.unit_solutions[m.fs.unit.area] = 26.63124

        comp_lst = ["NaCl", "H2O"]

        flow_mass_inlet = sum(
            m.fs.unit.feed_side.properties[0, 0].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_retentate = sum(
            m.fs.unit.feed_side.properties[0, 1].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_permeate = sum(
            m.fs.unit.mixed_permeate[0].flow_mass_phase_comp["Liq", j] for j in comp_lst
        )

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "NaCl"],
                "out": m.fs.unit.feed_side.properties[0, 0].flow_mass_phase_comp[
                    "Liq", "NaCl"
                ],
            },
            "Check 2": {
                "in": flow_mass_inlet,
                "out": flow_mass_retentate + flow_mass_permeate,
            },
        }

        return m


def build_basic():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = props.NaClParameterBlock()

    m.fs.unit = ReverseOsmosis1D(
        property_package=m.fs.properties,
        has_pressure_change=False,
        concentration_polarization_type=ConcentrationPolarizationType.none,
        mass_transfer_coefficient=MassTransferCoefficient.none,
        transformation_scheme="BACKWARD",
        transformation_method="dae.finite_difference",
        finite_elements=3,
    )

    # fully specify system
    feed_flow_mass = 1
    feed_mass_frac_NaCl = 0.035
    feed_pressure = 50e5
    feed_temperature = 273.15 + 25
    A = 4.2e-12
    B = 3.5e-8
    pressure_atmospheric = 101325
    feed_mass_frac_H2O = 1 - feed_mass_frac_NaCl

    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "NaCl"].fix(
        feed_flow_mass * feed_mass_frac_NaCl
    )

    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
        feed_flow_mass * feed_mass_frac_H2O
    )

    m.fs.unit.inlet.pressure[0].fix(feed_pressure)
    m.fs.unit.inlet.temperature[0].fix(feed_temperature)
    m.fs.unit.A_comp.fix(A)
    m.fs.unit.B_comp.fix(B)
    m.fs.unit.permeate.pressure[0].fix(pressure_atmospheric)
    m.fs.unit.length.fix(8)
    m.fs.unit.recovery_vol_phase[0, "Liq"].fix(0.4)

    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e1, index=("Liq", "H2O")
    )
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e3, index=("Liq", "NaCl")
    )

    iscale.calculate_scaling_factors(m.fs.unit)

    return m


class TestReverseOsmosis1D_basic(UnitTestHarness):
    def configure(self):
        m = build_basic()
        x_interface_in = m.fs.unit.feed_side.length_domain.at(2)

        self.unit_solutions[
            m.fs.unit.flux_mass_phase_comp[0, x_interface_in, "Liq", "H2O"]
        ] = 0.00484064
        self.unit_solutions[
            m.fs.unit.flux_mass_phase_comp[0, x_interface_in, "Liq", "NaCl"]
        ] = 1.628559e-6
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp[0, 1, "Liq", "H2O"]] = (
            0.001019483
        )
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp[0, 1, "Liq", "NaCl"]] = (
            1.99996e-6
        )
        self.unit_solutions[
            m.fs.unit.mixed_permeate[0].flow_mass_phase_comp["Liq", "H2O"]
        ] = 0.3895746
        self.unit_solutions[
            m.fs.unit.mixed_permeate[0].flow_mass_phase_comp["Liq", "NaCl"]
        ] = 0.000265183
        self.unit_solutions[m.fs.unit.area] = 144.307206

        comp_lst = ["NaCl", "H2O"]

        flow_mass_inlet = sum(
            m.fs.unit.feed_side.properties[0, 0].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_retentate = sum(
            m.fs.unit.feed_side.properties[0, 1].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_permeate = sum(
            m.fs.unit.mixed_permeate[0].flow_mass_phase_comp["Liq", j] for j in comp_lst
        )

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "NaCl"],
                "out": m.fs.unit.feed_side.properties[0, 0].flow_mass_phase_comp[
                    "Liq", "NaCl"
                ],
            },
            "Check 2": {
                "in": flow_mass_inlet,
                "out": flow_mass_retentate + flow_mass_permeate,
            },
        }

        return m


def build_SKK():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = props.NaClParameterBlock()

    m.fs.unit = ReverseOsmosis1D(
        property_package=m.fs.properties,
        has_pressure_change=True,
        concentration_polarization_type=ConcentrationPolarizationType.calculated,
        mass_transfer_coefficient=MassTransferCoefficient.calculated,
        pressure_change_type=PressureChangeType.calculated,
        transport_model=TransportModel.SKK,
        transformation_scheme="BACKWARD",
        transformation_method="dae.finite_difference",
        finite_elements=3,
        has_full_reporting=True,
    )

    # fully specify system
    feed_flow_mass = 1000 / 3600
    feed_mass_frac_NaCl = 0.034283
    feed_pressure = 70e5

    feed_temperature = 273.15 + 25
    A = 4.2e-12
    B = 3.5e-8
    pressure_atmospheric = 1e5
    feed_mass_frac_H2O = 1 - feed_mass_frac_NaCl

    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "NaCl"].fix(
        feed_flow_mass * feed_mass_frac_NaCl
    )

    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
        feed_flow_mass * feed_mass_frac_H2O
    )

    m.fs.unit.inlet.pressure[0].fix(feed_pressure)
    m.fs.unit.inlet.temperature[0].fix(feed_temperature)
    m.fs.unit.A_comp.fix(A)
    m.fs.unit.B_comp.fix(B)
    m.fs.unit.reflect_coeff.fix(0.9)
    m.fs.unit.permeate.pressure[0].fix(pressure_atmospheric)
    m.fs.unit.feed_side.N_Re[0, 0].fix(400)
    m.fs.unit.recovery_mass_phase_comp[0, "Liq", "H2O"].fix(0.5)
    m.fs.unit.feed_side.spacer_porosity.fix(0.97)
    m.fs.unit.feed_side.channel_height.fix(0.001)

    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e1, index=("Liq", "H2O")
    )
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e3, index=("Liq", "NaCl")
    )

    iscale.calculate_scaling_factors(m.fs.unit)

    return m


class TestReverseOsmosis1D_SKK(UnitTestHarness):
    def configure(self):
        m = build_SKK()
        x_interface_in = m.fs.unit.feed_side.length_domain.at(2)

        self.unit_solutions[m.fs.unit.deltaP[0]] = -98760.812
        self.unit_solutions[
            m.fs.unit.flux_mass_phase_comp[0, x_interface_in, "Liq", "H2O"]
        ] = 0.01225735
        self.unit_solutions[
            m.fs.unit.flux_mass_phase_comp[0, x_interface_in, "Liq", "NaCl"]
        ] = 7.4932174e-5
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp[0, 1, "Liq", "H2O"]] = (
            0.006690554
        )
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp[0, 1, "Liq", "NaCl"]] = (
            5.422075e-5
        )
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp_avg[0, "Liq", "H2O"]] = (
            0.0094392
        )
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp_avg[0, "Liq", "NaCl"]] = (
            6.5286415e-5
        )
        self.unit_solutions[
            m.fs.unit.mixed_permeate[0].flow_mass_phase_comp["Liq", "H2O"]
        ] = 0.1341274
        self.unit_solutions[
            m.fs.unit.mixed_permeate[0].flow_mass_phase_comp["Liq", "NaCl"]
        ] = 0.0009276945
        self.unit_solutions[m.fs.unit.feed_side.N_Re_avg[0]] = 383.229748
        self.unit_solutions[m.fs.unit.feed_side.K_avg[0, "NaCl"]] = 3.0276974e-5
        self.unit_solutions[m.fs.unit.area] = 14.20961

        comp_lst = ["NaCl", "H2O"]

        flow_mass_inlet = sum(
            m.fs.unit.feed_side.properties[0, 0].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_retentate = sum(
            m.fs.unit.feed_side.properties[0, 1].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_permeate = sum(
            m.fs.unit.mixed_permeate[0].flow_mass_phase_comp["Liq", j] for j in comp_lst
        )

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "NaCl"],
                "out": m.fs.unit.feed_side.properties[0, 0].flow_mass_phase_comp[
                    "Liq", "NaCl"
                ],
            },
            "Check 2": {
                "in": flow_mass_inlet,
                "out": flow_mass_retentate + flow_mass_permeate,
            },
        }

        return m


def build_cp_mod_fixed():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = props.NaClParameterBlock()

    m.fs.unit = ReverseOsmosis1D(
        property_package=m.fs.properties,
        has_pressure_change=False,
        concentration_polarization_type=ConcentrationPolarizationType.fixed,
        mass_transfer_coefficient=MassTransferCoefficient.none,
        transformation_scheme="BACKWARD",
        transformation_method="dae.finite_difference",
        finite_elements=3,
    )

    # fully specify system
    feed_flow_mass = 1
    feed_mass_frac_NaCl = 0.035
    feed_pressure = 50e5
    feed_temperature = 273.15 + 25
    A = 4.2e-12
    B = 3.5e-8
    pressure_atmospheric = 101325
    feed_mass_frac_H2O = 1 - feed_mass_frac_NaCl

    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "NaCl"].fix(
        feed_flow_mass * feed_mass_frac_NaCl
    )

    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
        feed_flow_mass * feed_mass_frac_H2O
    )

    m.fs.unit.inlet.pressure[0].fix(feed_pressure)
    m.fs.unit.inlet.temperature[0].fix(feed_temperature)
    m.fs.unit.A_comp.fix(A)
    m.fs.unit.B_comp.fix(B)
    m.fs.unit.permeate.pressure[0].fix(pressure_atmospheric)
    m.fs.unit.length.fix(8)
    m.fs.unit.recovery_vol_phase[0, "Liq"].fix(0.4)
    m.fs.unit.feed_side.cp_modulus.fix(1.1)

    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e1, index=("Liq", "H2O")
    )
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e3, index=("Liq", "NaCl")
    )

    iscale.calculate_scaling_factors(m.fs.unit)

    return m


class TestReverseOsmosis1D_cp_mod_fixed(UnitTestHarness):
    def configure(self):
        m = build_cp_mod_fixed()
        x_interface_in = m.fs.unit.feed_side.length_domain.at(2)

        self.unit_solutions[
            m.fs.unit.flux_mass_phase_comp[0, x_interface_in, "Liq", "H2O"]
        ] = 0.0024494142
        self.unit_solutions[
            m.fs.unit.flux_mass_phase_comp[0, x_interface_in, "Liq", "NaCl"]
        ] = 1.863938e-6
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp[0, 1, "Liq", "H2O"]] = (
            0.000359974
        )
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp[0, 1, "Liq", "NaCl"]] = (
            2.051563e-6
        )
        self.unit_solutions[
            m.fs.unit.mixed_permeate[0].flow_mass_phase_comp["Liq", "H2O"]
        ] = 0.3894809
        self.unit_solutions[
            m.fs.unit.mixed_permeate[0].flow_mass_phase_comp["Liq", "NaCl"]
        ] = 0.000652886
        self.unit_solutions[m.fs.unit.area] = 329.708005

        comp_lst = ["NaCl", "H2O"]

        flow_mass_inlet = sum(
            m.fs.unit.feed_side.properties[0, 0].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_retentate = sum(
            m.fs.unit.feed_side.properties[0, 1].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_permeate = sum(
            m.fs.unit.mixed_permeate[0].flow_mass_phase_comp["Liq", j] for j in comp_lst
        )

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "NaCl"],
                "out": m.fs.unit.feed_side.properties[0, 0].flow_mass_phase_comp[
                    "Liq", "NaCl"
                ],
            },
            "Check 2": {
                "in": flow_mass_inlet,
                "out": flow_mass_retentate + flow_mass_permeate,
            },
        }

        return m


def build_cp_calculated_kf_fixed():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = props.NaClParameterBlock()

    m.fs.unit = ReverseOsmosis1D(
        property_package=m.fs.properties,
        has_pressure_change=False,
        concentration_polarization_type=ConcentrationPolarizationType.calculated,
        mass_transfer_coefficient=MassTransferCoefficient.fixed,
        transformation_scheme="BACKWARD",
        transformation_method="dae.finite_difference",
        finite_elements=3,
    )

    # fully specify system
    feed_flow_mass = 1
    feed_mass_frac_NaCl = 0.035
    feed_pressure = 50e5
    feed_temperature = 273.15 + 25
    A = 4.2e-12
    B = 3.5e-8
    pressure_atmospheric = 101325
    feed_mass_frac_H2O = 1 - feed_mass_frac_NaCl

    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "NaCl"].fix(
        feed_flow_mass * feed_mass_frac_NaCl
    )

    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
        feed_flow_mass * feed_mass_frac_H2O
    )

    m.fs.unit.inlet.pressure[0].fix(feed_pressure)
    m.fs.unit.inlet.temperature[0].fix(feed_temperature)
    # m.fs.unit.dP_dx.fix(0)
    m.fs.unit.A_comp.fix(A)
    m.fs.unit.B_comp.fix(B)
    m.fs.unit.permeate.pressure[0].fix(pressure_atmospheric)
    m.fs.unit.length.fix(8)
    m.fs.unit.recovery_vol_phase[0, "Liq"].fix(0.4)
    m.fs.unit.feed_side.K.fix(2e-5)

    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e1, index=("Liq", "H2O")
    )
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e3, index=("Liq", "NaCl")
    )

    iscale.calculate_scaling_factors(m.fs.unit)

    return m


class TestReverseOsmosis1D_cp_calculated_kf_fixed(UnitTestHarness):
    def configure(self):
        m = build_cp_calculated_kf_fixed()
        x_interface_in = m.fs.unit.feed_side.length_domain.at(2)

        self.unit_solutions[
            m.fs.unit.flux_mass_phase_comp[0, x_interface_in, "Liq", "H2O"]
        ] = 0.002791588
        self.unit_solutions[
            m.fs.unit.flux_mass_phase_comp[0, x_interface_in, "Liq", "NaCl"]
        ] = 1.83061e-6
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp[0, 1, "Liq", "H2O"]] = (
            0.0007081512
        )
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp[0, 1, "Liq", "NaCl"]] = (
            2.027294e-6
        )
        self.unit_solutions[
            m.fs.unit.mixed_permeate[0].flow_mass_phase_comp["Liq", "H2O"]
        ] = 0.3895265
        self.unit_solutions[
            m.fs.unit.mixed_permeate[0].flow_mass_phase_comp["Liq", "NaCl"]
        ] = 0.000464535

        comp_lst = ["NaCl", "H2O"]

        flow_mass_inlet = sum(
            m.fs.unit.feed_side.properties[0, 0].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_retentate = sum(
            m.fs.unit.feed_side.properties[0, 1].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_permeate = sum(
            m.fs.unit.mixed_permeate[0].flow_mass_phase_comp["Liq", j] for j in comp_lst
        )

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "NaCl"],
                "out": m.fs.unit.feed_side.properties[0, 0].flow_mass_phase_comp[
                    "Liq", "NaCl"
                ],
            },
            "Check 2": {
                "in": flow_mass_inlet,
                "out": flow_mass_retentate + flow_mass_permeate,
            },
        }

        return m


def build_cp_calculated_kf_calculated():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = props.NaClParameterBlock()

    m.fs.unit = ReverseOsmosis1D(
        property_package=m.fs.properties,
        has_pressure_change=False,
        concentration_polarization_type=ConcentrationPolarizationType.calculated,
        mass_transfer_coefficient=MassTransferCoefficient.calculated,
        transformation_scheme="BACKWARD",
        transformation_method="dae.finite_difference",
        finite_elements=3,
    )

    # fully specify system
    feed_flow_mass = 1
    feed_mass_frac_NaCl = 0.035
    feed_pressure = 50e5
    feed_temperature = 273.15 + 25
    A = 4.2e-12
    B = 3.5e-8
    pressure_atmospheric = 101325
    feed_mass_frac_H2O = 1 - feed_mass_frac_NaCl

    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "NaCl"].fix(
        feed_flow_mass * feed_mass_frac_NaCl
    )

    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
        feed_flow_mass * feed_mass_frac_H2O
    )

    m.fs.unit.inlet.pressure[0].fix(feed_pressure)
    m.fs.unit.inlet.temperature[0].fix(feed_temperature)
    m.fs.unit.A_comp.fix(A)
    m.fs.unit.B_comp.fix(B)
    m.fs.unit.permeate.pressure[0].fix(pressure_atmospheric)
    m.fs.unit.length.fix(8)
    m.fs.unit.recovery_vol_phase[0, "Liq"].fix(0.4)
    m.fs.unit.feed_side.spacer_porosity.fix(0.75)
    m.fs.unit.feed_side.channel_height.fix(0.002)

    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e1, index=("Liq", "H2O")
    )
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e3, index=("Liq", "NaCl")
    )

    iscale.calculate_scaling_factors(m.fs.unit)

    return m


class TestReverseOsmosis1D_cp_calculated_kf_calculated(UnitTestHarness):
    def configure(self):
        m = build_cp_calculated_kf_calculated()
        x_interface_in = m.fs.unit.feed_side.length_domain.at(2)

        self.unit_solutions[
            m.fs.unit.flux_mass_phase_comp[0, x_interface_in, "Liq", "H2O"]
        ] = 0.002382908
        self.unit_solutions[
            m.fs.unit.flux_mass_phase_comp[0, x_interface_in, "Liq", "NaCl"]
        ] = 1.8703962e-6
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp[0, 1, "Liq", "H2O"]] = (
            0.000626977
        )
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp[0, 1, "Liq", "NaCl"]] = (
            2.0339201e-6
        )
        self.unit_solutions[
            m.fs.unit.mixed_permeate[0].flow_mass_phase_comp["Liq", "H2O"]
        ] = 0.3895066
        self.unit_solutions[
            m.fs.unit.mixed_permeate[0].flow_mass_phase_comp["Liq", "NaCl"]
        ] = 0.000546681

        comp_lst = ["NaCl", "H2O"]

        flow_mass_inlet = sum(
            m.fs.unit.feed_side.properties[0, 0].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_retentate = sum(
            m.fs.unit.feed_side.properties[0, 1].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_permeate = sum(
            m.fs.unit.mixed_permeate[0].flow_mass_phase_comp["Liq", j] for j in comp_lst
        )

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "NaCl"],
                "out": m.fs.unit.feed_side.properties[0, 0].flow_mass_phase_comp[
                    "Liq", "NaCl"
                ],
            },
            "Check 2": {
                "in": flow_mass_inlet,
                "out": flow_mass_retentate + flow_mass_permeate,
            },
        }

        return m


def build_friction_factor_spiral_wound():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = props.NaClParameterBlock()

    m.fs.unit = ReverseOsmosis1D(
        property_package=m.fs.properties,
        has_pressure_change=True,
        concentration_polarization_type=ConcentrationPolarizationType.calculated,
        mass_transfer_coefficient=MassTransferCoefficient.calculated,
        pressure_change_type=PressureChangeType.calculated,
        module_type=ModuleType.spiral_wound,
        transformation_scheme="BACKWARD",
        transformation_method="dae.finite_difference",
        finite_elements=3,
    )

    # fully specify system
    feed_flow_mass = 1
    feed_mass_frac_NaCl = 0.035
    feed_pressure = 50e5
    feed_temperature = 273.15 + 25
    membrane_area = 19
    A = 4.2e-12
    B = 3.5e-8
    pressure_atmospheric = 101325
    feed_mass_frac_H2O = 1 - feed_mass_frac_NaCl

    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "NaCl"].fix(
        feed_flow_mass * feed_mass_frac_NaCl
    )

    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
        feed_flow_mass * feed_mass_frac_H2O
    )

    m.fs.unit.inlet.pressure[0].fix(feed_pressure)
    m.fs.unit.inlet.temperature[0].fix(feed_temperature)
    m.fs.unit.area.fix(membrane_area)
    m.fs.unit.A_comp.fix(A)
    m.fs.unit.B_comp.fix(B)
    m.fs.unit.permeate.pressure[0].fix(pressure_atmospheric)
    m.fs.unit.length.fix(8)
    m.fs.unit.feed_side.spacer_porosity.fix(0.75)
    m.fs.unit.feed_side.channel_height.fix(0.002)

    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e1, index=("Liq", "H2O")
    )
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e3, index=("Liq", "NaCl")
    )
    iscale.calculate_scaling_factors(m.fs.unit)

    return m


class TestReverseOsmosis1D_friction_factor_spiral_wound(UnitTestHarness):
    def configure(self):
        m = build_friction_factor_spiral_wound()
        x_interface_in = m.fs.unit.feed_side.length_domain.at(2)

        self.unit_solutions[
            m.fs.unit.flux_mass_phase_comp[0, x_interface_in, "Liq", "H2O"]
        ] = 0.00593246
        self.unit_solutions[
            m.fs.unit.flux_mass_phase_comp[0, x_interface_in, "Liq", "NaCl"]
        ] = 1.4481768e-6
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp[0, 1, "Liq", "H2O"]] = (
            0.0041822435
        )
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp[0, 1, "Liq", "NaCl"]] = (
            1.491610924e-6
        )
        self.unit_solutions[
            m.fs.unit.mixed_permeate[0].flow_mass_phase_comp["Liq", "H2O"]
        ] = 0.095819393
        self.unit_solutions[
            m.fs.unit.mixed_permeate[0].flow_mass_phase_comp["Liq", "NaCl"]
        ] = 2.794556261e-5

        comp_lst = ["NaCl", "H2O"]

        flow_mass_inlet = sum(
            m.fs.unit.feed_side.properties[0, 0].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_retentate = sum(
            m.fs.unit.feed_side.properties[0, 1].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_permeate = sum(
            m.fs.unit.mixed_permeate[0].flow_mass_phase_comp["Liq", j] for j in comp_lst
        )

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "NaCl"],
                "out": m.fs.unit.feed_side.properties[0, 0].flow_mass_phase_comp[
                    "Liq", "NaCl"
                ],
            },
            "Check 2": {
                "in": flow_mass_inlet,
                "out": flow_mass_retentate + flow_mass_permeate,
            },
        }

        return m


def build_cp_calculated_kf_calculated_pdrop_fixed_by_dx():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = props.NaClParameterBlock()

    m.fs.unit = ReverseOsmosis1D(
        property_package=m.fs.properties,
        has_pressure_change=True,
        concentration_polarization_type=ConcentrationPolarizationType.calculated,
        mass_transfer_coefficient=MassTransferCoefficient.calculated,
        pressure_change_type=PressureChangeType.fixed_per_unit_length,
        transformation_scheme="BACKWARD",
        transformation_method="dae.finite_difference",
        finite_elements=3,
    )

    # fully specify system
    feed_flow_mass = 1
    feed_mass_frac_NaCl = 0.035
    feed_pressure = 50e5
    feed_temperature = 273.15 + 25
    A = 4.2e-12
    B = 3.5e-8
    pressure_atmospheric = 101325
    feed_mass_frac_H2O = 1 - feed_mass_frac_NaCl

    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "NaCl"].fix(
        feed_flow_mass * feed_mass_frac_NaCl
    )

    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
        feed_flow_mass * feed_mass_frac_H2O
    )

    m.fs.unit.inlet.pressure[0].fix(feed_pressure)
    m.fs.unit.inlet.temperature[0].fix(feed_temperature)
    m.fs.unit.A_comp.fix(A)
    m.fs.unit.B_comp.fix(B)
    m.fs.unit.permeate.pressure[0].fix(pressure_atmospheric)
    m.fs.unit.length.fix(8)
    m.fs.unit.recovery_vol_phase[0, "Liq"].fix(0.4)
    m.fs.unit.feed_side.spacer_porosity.fix(0.75)
    m.fs.unit.feed_side.channel_height.fix(0.002)
    m.fs.unit.feed_side.dP_dx.fix(-0.1e5)

    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e1, index=("Liq", "H2O")
    )
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e3, index=("Liq", "NaCl")
    )

    iscale.calculate_scaling_factors(m.fs.unit)

    return m


class TestReverseOsmosis1D_cp_calculated_kf_calculated_pdrop_fixed_by_dx(
    UnitTestHarness
):
    def configure(self):
        m = build_cp_calculated_kf_calculated_pdrop_fixed_by_dx()
        x_interface_in = m.fs.unit.feed_side.length_domain.at(2)

        self.unit_solutions[
            m.fs.unit.flux_mass_phase_comp[0, x_interface_in, "Liq", "H2O"]
        ] = 0.00224943
        self.unit_solutions[
            m.fs.unit.flux_mass_phase_comp[0, x_interface_in, "Liq", "NaCl"]
        ] = 1.872372e-6
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp[0, 1, "Liq", "H2O"]] = (
            0.000551092
        )
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp[0, 1, "Liq", "NaCl"]] = (
            2.0075516e-6
        )
        self.unit_solutions[
            m.fs.unit.mixed_permeate[0].flow_mass_phase_comp["Liq", "H2O"]
        ] = 0.3894964
        self.unit_solutions[
            m.fs.unit.mixed_permeate[0].flow_mass_phase_comp["Liq", "NaCl"]
        ] = 0.000588787

        comp_lst = ["NaCl", "H2O"]

        flow_mass_inlet = sum(
            m.fs.unit.feed_side.properties[0, 0].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_retentate = sum(
            m.fs.unit.feed_side.properties[0, 1].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_permeate = sum(
            m.fs.unit.mixed_permeate[0].flow_mass_phase_comp["Liq", j] for j in comp_lst
        )

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "NaCl"],
                "out": m.fs.unit.feed_side.properties[0, 0].flow_mass_phase_comp[
                    "Liq", "NaCl"
                ],
            },
            "Check 2": {
                "in": flow_mass_inlet,
                "out": flow_mass_retentate + flow_mass_permeate,
            },
        }

        return m


def build_cp_calculated_kf_calculated_pdrop_fixed_by_stage():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = props.NaClParameterBlock()

    m.fs.unit = ReverseOsmosis1D(
        property_package=m.fs.properties,
        has_pressure_change=True,
        concentration_polarization_type=ConcentrationPolarizationType.calculated,
        mass_transfer_coefficient=MassTransferCoefficient.calculated,
        pressure_change_type=PressureChangeType.fixed_per_stage,
        transformation_scheme="BACKWARD",
        transformation_method="dae.finite_difference",
        finite_elements=3,
    )

    # fully specify system
    feed_flow_mass = 1
    feed_mass_frac_NaCl = 0.035
    feed_pressure = 50e5
    feed_temperature = 273.15 + 25
    A = 4.2e-12
    B = 3.5e-8
    pressure_atmospheric = 101325
    feed_mass_frac_H2O = 1 - feed_mass_frac_NaCl

    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "NaCl"].fix(
        feed_flow_mass * feed_mass_frac_NaCl
    )

    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
        feed_flow_mass * feed_mass_frac_H2O
    )

    m.fs.unit.inlet.pressure[0].fix(feed_pressure)
    m.fs.unit.inlet.temperature[0].fix(feed_temperature)
    m.fs.unit.A_comp.fix(A)
    m.fs.unit.B_comp.fix(B)
    m.fs.unit.permeate.pressure[0].fix(pressure_atmospheric)
    m.fs.unit.length.fix(8)
    m.fs.unit.recovery_vol_phase[0, "Liq"].fix(0.4)
    m.fs.unit.feed_side.spacer_porosity.fix(0.75)
    m.fs.unit.feed_side.channel_height.fix(0.002)
    m.fs.unit.deltaP.fix(-62435.6)

    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e1, index=("Liq", "H2O")
    )
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e3, index=("Liq", "NaCl")
    )

    iscale.calculate_scaling_factors(m.fs.unit)

    return m


class TestReverseOsmosis1D_cp_calculated_kf_calculated_pdrop_fixed_by_stage(
    UnitTestHarness
):
    def configure(self):
        m = build_cp_calculated_kf_calculated_pdrop_fixed_by_stage()

        x_interface_in = m.fs.unit.feed_side.length_domain.at(2)

        self.unit_solutions[
            m.fs.unit.flux_mass_phase_comp[0, x_interface_in, "Liq", "H2O"]
        ] = 0.002278454
        self.unit_solutions[
            m.fs.unit.flux_mass_phase_comp[0, x_interface_in, "Liq", "NaCl"]
        ] = 1.871967e-6
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp[0, 1, "Liq", "H2O"]] = (
            0.000566998
        )
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp[0, 1, "Liq", "NaCl"]] = (
            2.0134278e-6
        )
        self.unit_solutions[
            m.fs.unit.mixed_permeate[0].flow_mass_phase_comp["Liq", "H2O"]
        ] = 0.3894987
        self.unit_solutions[
            m.fs.unit.mixed_permeate[0].flow_mass_phase_comp["Liq", "NaCl"]
        ] = 0.0005792222

        comp_lst = ["NaCl", "H2O"]

        flow_mass_inlet = sum(
            m.fs.unit.feed_side.properties[0, 0].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_retentate = sum(
            m.fs.unit.feed_side.properties[0, 1].flow_mass_phase_comp["Liq", j]
            for j in comp_lst
        )
        flow_mass_permeate = sum(
            m.fs.unit.mixed_permeate[0].flow_mass_phase_comp["Liq", j] for j in comp_lst
        )

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "NaCl"],
                "out": m.fs.unit.feed_side.properties[0, 0].flow_mass_phase_comp[
                    "Liq", "NaCl"
                ],
            },
            "Check 2": {
                "in": flow_mass_inlet,
                "out": flow_mass_retentate + flow_mass_permeate,
            },
        }

        return m
